"""Implements the DCM Cloud API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camelize, camel_keys, uncamel_keys


class Subscription(Resource):

    """ A subscription describes the capabilities of a specific region as matched by your subscription to
    the region. You can use this object to tell what cloud capabilities exist in this region. In some
    cases, failure to support something will be a limitation/design choice of the cloud in question. In
    other cases, it just represents the fact that you have yet to subscribe to the service in question. """
    PATH = 'geography/Subscription'
    COLLECTION_NAME = 'subscriptions'
    PRIMARY_KEY = 'region_id'

    def __init__(self, region_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__region_id = region_id

    @classmethod
    def all(cls, keys_only=False,endpoint=None, **kwargs):
        if 'region_id' in kwargs:
            r = Resource(cls.PATH + "/" + str(kwargs['region_id']), endpoint=endpoint)
        else:
            r = Resource(cls.PATH, endpoint=endpoint)

        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        x = r.get()
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise SubscriptionException(r.last_error)


class SubscriptionException(BaseException):
    pass



"""Implements the DCM Network API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
from mixcoatl.admin.job import Job
import json


class Network(Resource):

    """ Dell Cloud Manager models two distinct kinds of networks as network resources:

    * Standard networks such as an AWS VPC or Cloud.com network that represent a 
      network as known to a cloud provider
    * Overlay networks such as a VPNCubed, CloudSwitch, or vCider network in which 
      the network is an overlay on top of the cloud providers network 
    """
    PATH = 'network/Network'
    COLLECTION_NAME = 'networks'
    PRIMARY_KEY = "network_id"

    def __init__(self, network_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)

        if 'detail' in kwargs:
            self.request_details = kwargs['detail']

        self.__network_id = network_id

    @property
    def network_id(self):
        """`str` - The unique ID for this network"""
        return self.__network_id

    @lazy_property
    def status(self):
        """`str` - The status of this network"""
        return self.__status

    @lazy_property
    def label(self):
        """`str` - A color label assigned to this network"""
        return self.__label

    @lazy_property
    def data_center(self):
        return self.__data_center

    @label.setter
    def label(self, l):
        self.__label = l

    @lazy_property
    def name(self):
        """`str` - The user-friendly name for this network"""
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n

    @lazy_property
    def description(self):
        """`str` - The description of this network"""
        return self.__description

    @description.setter
    def description(self, d):
        self.__description = d

    @lazy_property
    def account(self):
        """`dict` - The account associated with the network"""
        return self.__account

    @lazy_property
    def owning_groups(self):
        """`list` - The DCM groups that have ownership of this network"""
        return self.__owning_groups

    @owning_groups.setter
    def owning_groups(self, g):
        self.__owning_groups = [{'group_id': g}]

    @lazy_property
    def cloud(self):
        """`dict` - The DCM cloud account in which this network lives"""
        return self.__cloud

    @lazy_property
    def data_center(self):
        """`dict` - The data center (if any) to which this network is tied"""
        return slef.__data_center

    @lazy_property
    def provider_id(self):
        """`str` - The cloud provider unique id for this network"""
        return self.__provider_id

    @lazy_property
    def region(self):
        """`dict` - The region in which this network operates"""
        return self.__region

    @region.setter
    def region(self, r):
        self.__region = {'region_id': r}

    @lazy_property
    def removable(self):
        """`bool` - Indicates if this network can be removed"""
        return self.__removable

    @lazy_property
    def budget(self):
        """`int` - The DCM billing code costs are associated with"""
        return self.__budget

    @budget.setter
    def budget(self, b):
        self.__budget = b

    @lazy_property
    def customer(self):
        """`dict` - The DCM customer to which this network belongs"""
        return self.__customer

    @lazy_property
    def guid(self):
        """`str` - The permanent unique URI identifier for this network in DCM."""
        return self.__guid

    @lazy_property
    def agent_communication(self):
        """`bool` - Indicates whether communication between DCM and the agents on the guest operating systems 
        in the cloud occur over a public or private channel."""
        return self.__agent_communication

    @lazy_property
    def allow_subnet_creation(self):
        """`bool` - Indicates whether or not you can POST to the Subnet resource to dynamically create subnets 
        against this network."""
        return self.__allow_subnet_creation

    @lazy_property
    def can_have_subnets(self):
        """`bool` - Indicates whether the technology behind this network supports subdivision into subnets."""
        return self.__can_have_subnets

    @lazy_property
    def publicly_addressable(self):
        """`bool` - Indicates whether the resources in this network can be accessed from the public Internet."""
        return self.__publicly_addressable

    @lazy_property
    def flat(self):
        """`bool` - Indicates whether this network is a flat network or is part of a more fine-grained 
        network hierarchy."""
        return self.__flat

    @lazy_property
    def created_timestamp(self):
        """`str` - Creation time."""
        return self.__created_timestamp

    @lazy_property
    def last_modified_timestamp(self):
        """`str` - The time when a modification was last made to this network."""
        return self.__last_modified_timestamp

    @lazy_property
    def network_address(self):
        """`str` - An IPv4 CIDR representing the block of IP addresses hosted in this network."""
        return self.__network_address

    @network_address.setter
    def network_address(self, n):
        self.__network_address = n

    @lazy_property
    def network_type(self):
        """`str` - Identifies what kind of network is represented by this resource."""
        return self.__network_type

    @lazy_property
    def ntp_servers(self):
        """`list` - A list of NTP servers configured for this network."""
        return self.__ntp_servers

    @ntp_servers.setter
    def ntp_servers(self, n):
        self.__ntp_servers = [n]

    @lazy_property
    def subnets(self):
        """`list` - The list of subnets associated with this network."""
        return self.__subnets

    @lazy_property
    def dns_servers(self):
        """`list` - A list of DNS servers configured for this network."""
        return self.__dns_servers

    @dns_servers.setter
    def dns_servers(self, d):
        self.__dns_servers = [d]

    @lazy_property
    def owning_user(self):
        """`str` - The network's owning user. """
        return self.__owning_user

    @required_attrs(['budget', 'name', 'network_address', 'region', 'description'])
    def create(self, **kwargs):
        """Create a new network

        :param callback: Optional callback to call with resulting :class:`Network`
        :type callback: func.
        :returns: :class:`Job`
        :raises: :class:`NetworkException`
        """

        optional_attrs = [
            'owning_groups', 'ntp_servers', 'dns_servers', 'label']
        payload = {'add_network': [{
                   'budget': self.budget,
                   'name': self.name,
                   'network_address': self.network_address,
                   'description': self.description,
                   'region': self.region}]}

        if 'label' in kwargs:
            payload['add_network'][0]['label'] = kwargs['label']

        callback = kwargs.get('callback', None)

        for oa in optional_attrs:
            try:
                if getattr(self, oa) is not None:
                    payload['add_network'][0].update({oa: getattr(self, oa)})
            except AttributeError:
                pass

        self.post(self.PATH, data=json.dumps(camel_keys(payload)))

        if self.last_error is None:
            j = Job(self.current_job, endpoint=self.endpoint)
            j.load()
            if callback is not None:
                callback(j)
            else:
                return j
        else:
            raise NetworkException(self.last_error)

    @required_attrs(['network_id'])
    def destroy(self, reason='no reason provided'):
        """Destroy network with reason :attr:`reason`

        :param reason: The reason for terminating the server
        :type reason: str.
        :returns: bool -- Result of API call
        """
        p = self.PATH + "/" + str(self.network_id)
        qopts = {'reason': reason}
        return self.delete(p, params=qopts)

    @classmethod
    def all(cls, endpoint=None, **kwargs):
        """List all networks in `region_id`

        :param region_id: Limit results to `region_id`
        :type region_id: int.
        :param data_center_id: Limit results to `data_center_id`
        :type region_id: int.
        :param account_id: limit results to `account_id`
        :type account_id: int.
        :param detail: Level of detail to return - `basic` or `extended`
        :type detail: str.
        :param keys_only: Return only :attr:`network_id` in results
        :type keys_only: bool.
        :param active_only: Limits the list of networks to only active networks if true. Default is True.
        :type keys_only: bool.
        :returns: `list` of :attr:`network_id` or :class:`Network`
        :raises: :class:`NetworkException`
        """
        params = {}
        r = Resource(cls.PATH,endpoint=endpoint)

        if 'detail' in kwargs:
            request_details = kwargs['detail']
        else:
            request_details = 'extended'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        if 'region_id' in kwargs:
            params['regionId'] = kwargs['region_id']

        if 'data_center_id' in kwargs:
            params['dataCenterId'] = kwargs['data_center_id']

        if 'account_id' in kwargs:
            params['accountId'] = kwargs['account_id']

        if 'active_only' in kwargs:
            params['activeOnly'] = kwargs['active_only']
        else:
            params['activeOnly'] = True

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise NetworkException(r.last_error)


class NetworkException(BaseException):

    """Generic Network Exception"""
    pass



"""Implements the DCM Group API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
from mixcoatl.infrastructure.snapshot import Snapshot
from mixcoatl.infrastructure.snapshot import SnapshotException
from mixcoatl.admin.job import Job
import json
import time


class Group(Resource):

    """A group is a collection of users. Groups map to roles that define what their access rights are for
    a specific account. Some operations may not be possible if your groups are being managed in
    LDAP or ActiveDirectory as Dell Cloud Manager is not the authority for that data in those
    instances. """
    PATH = 'admin/Group'
    COLLECTION_NAME = 'groups'
    PRIMARY_KEY = 'group_id'

    def __init__(self, group_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__group_id = group_id

    @property
    def group_id(self):
        """`int` - The unique id of this group in DCM"""
        return self.__group_id

    @lazy_property
    def description(self):
        """`str` - The user-friendly description of this group"""
        return self.__description

    @description.setter
    def description(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__description = b

    @lazy_property
    def name(self):
        """`str` - The name of the group"""
        return self.__name

    @name.setter
    def name(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__name = b

    @lazy_property
    def status(self):
        """`str` - The current status of the group in DCM"""
        return self.__status

    @lazy_property
    def customer(self):
        """`dict` - The customer to who this group belongs."""
        return self.__customer

    @lazy_property
    def customer_admin(self):
        """`bool` """
        return self.__customer_admin

    @lazy_property
    def user_admin(self):
        """`bool` """
        return self.__user_admin

    @lazy_property
    def configuration_management_admin(self):
        """`bool` """
        return self.__configuration_management_admin

    @lazy_property
    def configuration_management_qualifier(self):
        """`bool` """
        return self.__configuration_management_qualifier

    @lazy_property
    def group_account_budget_codes(self):
        """`list` """
        return self.__group_account_budget_codes

    @lazy_property
    def role_assignments(self):
        """`list` """
        return self.__role_assignments

    @classmethod
    def all(cls, keys_only=False, endpoint=None, **kwargs):
        """Get all groups

        .. note::

            The keys used to make the request determine results visibility

        :param keys_only: Only return `group_id` instead of `Group` objects
        :type keys_only: bool.
        :param detail: The level of detail to return - `basic` or `extended`
        :type detail: str.
        :param account_id: Restrict results to `account_id`
        :type account_id: int.
        :returns: `list` - List of :class:`Group` or :attr:`group_id`
        :raises: :class:`GroupException`
        """
        r = Resource(cls.PATH, endpoint=endpoint)
        params = {}

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'account_id' in kwargs:
            params = {'accountId': kwargs['account_id']}

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)]
                        for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i)
                        for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise GroupException(r.last_error)

    @required_attrs(['group_id'])
    def set_role(self, role_id, account_id):
        """Updates the role applied to the group specified by group_id.
           account_id is technically optional, but we're making it required
           here because making it optional may result in user problems.

           It's complicated.

           If you don't specify an account then it'll find an account associated with the billing
           credentials for the given user."""

        p = '%s/%s' % (self.PATH, str(self.group_id))
        payload = {'setRole': {
            'group': {'role': {'roleId': role_id}, 'account': {'accountId': account_id}}}}

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['group_id', 'name', 'description'])
    def update(self, name, description):
        """Updates the group name and description."""
        p = '%s/%s' % (self.PATH, str(self.group_id))
        payload = {'describeGroup': {
            'group': {'name': str(name), 'description': str(description)}}}

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['group_id', 'name'])
    def update_name(self, name):
        """Updates the group name. An update to one parameter requires both arguments."""
        p = '%s/%s' % (self.PATH, str(self.group_id))
        payload = {'describeGroup': {
            'group': {'name': str(name), 'description': self.description}}}

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['group_id', 'description'])
    def update_description(self, description):
        """Updates the group description. An update to one parameter requires both arguments."""
        p = '%s/%s' % (self.PATH, str(self.group_id))
        payload = {'describeGroup': {
            'group': {'name': self.name, 'description': str(description)}}}

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['name', 'description'])
    def create(self, callback=None):
        """Creates a new group

        :param callback: Optional callback to send the resulting :class:`Job`
        :raises: :class:`GroupCreationException`
        """
        parms = {'group': {'name': self.name,
                           'description': self.description}}
        payload = {'addGroup': camel_keys(parms)}

        response = self.post(data=json.dumps(payload))
        if self.last_error is None:
            self.load()
            return response
        else:
            raise GroupCreationException(self.last_error)


class GroupException(BaseException):
    pass


class GroupCreationException(GroupException):

    """Group Creation Exception"""
    pass



"""
mixcoatl.admin.api_key
----------------------

Implements access to the DCM ApiKey API
"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import uncamel, camelize, camel_keys, uncamel_keys
import json


class ApiKey(Resource):

    """An API key is an access key and secret key that provide API access into DCM."""
    PATH = 'admin/ApiKey'
    COLLECTION_NAME = 'apiKeys'
    PRIMARY_KEY = 'access_key'

    def __init__(self, access_key=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__access_key = access_key

    @property
    def access_key(self):
        """The primary identifier of the `ApiKey`. Same as `DCM_ACCESS_KEY`"""
        return self.__access_key

    @lazy_property
    def account(self):
        """`dict` - The account with which this API key is associated."""
        return self.__account

    @lazy_property
    def activation(self):
        """`str` - The date and time when this key was activated."""
        return self.__activation

    @lazy_property
    def expiration(self):
        """`str` - The date and time when this API key should automatically be made inactivate."""
        return self.__expiration

    @expiration.setter
    def expiration(self, e):
        self.__expiration = e

    @lazy_property
    def customer(self):
        """`dict` - The customer to whom this API key belongs."""
        return self.__customer

    @lazy_property
    def customer_management_key(self):
        """`bool` - Identifies whether or not this key can be used across all customer accounts."""
        return self.__customer_management_key

    @lazy_property
    def description(self):
        """`str` - A user-friendly description of this API key."""
        return self.__description

    @description.setter
    def description(self, d):
        self.__description = d

    @lazy_property
    def name(self):
        """`str` - The user-friendly name used to identify the key."""
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n

    @lazy_property
    def secret_key(self):
        """`str` - The secret part of this API key."""
        return self.__secret_key

    @lazy_property
    def state(self):
        """`str` - The status of the key *(i.e. `ACTIVE`)*"""
        return self.__state

    @lazy_property
    def system_management_key(self):
        """`bool` - Identifies if the key can be used for DCM system management functions"""
        return self.__system_management_key

    @lazy_property
    def user(self):
        """`dict` - The user associated with this API key. Account-level keys return `{'user_id': -1}`"""
        return self.__user

    @required_attrs(['description', 'name'])
    def create(self):
        """Call the API to generate an API key from the current instance of `ApiKey`"""

        payload = {
            'generateApiKey': [{'description': self.description, 'name': self.name}]}
        s = self.post(data=json.dumps(payload))
        if self.last_error is None:
            self.__access_key = s['apiKeys'][0]['accessKey']
            self.load()
        else:
            raise ApiKeyGenerationException(self.last_error)

    def invalidate(self, reason='key deleted via mixcoatl'):
        """Call the API to invalidate the current instance of `ApiKey`
        This is the same as deleting the api key

        :param reason: the reason for invalidating the key
        :type reason: str.
        :returns: True
        :raises: :class:`ApiKeyInvalidationException`
        """
        params = {'reason': reason}
        self.delete(params=params)
        if self.last_error is None:
            return True
        else:
            raise ApiKeyInvalidationException(self.last_error)

    @classmethod
    def generate_api_key(cls, key_name, description, expiration=None):
        """Generates a new API key

        >>> ApiKey.generate_api_key('my-api-key', 'this is my api key')
        {'access_key':'ABCDEFGHIJKL':....}

        :param key_name: the name for the key
        :type key_name: str.
        :param description: the description for the key
        :type description: str.
        :param expiration: *unused for now*
        :type expiration: str.
        :returns: :class:`ApiKey`
        :raises: :class:`ApiKeyGenerationException`
        """

        a = cls()
        a.name = key_name
        a.description = description
        a.create()
        return a

    @classmethod
    def all(cls, keys_only=False, endpoint=None, **kwargs):
        """Get all api keys

        .. note::

            The keys used to make the request determine results visibility

        :param keys_only: Only return `access_key` instead of `ApiKey` objects
        :type keys_only: bool.
        :param detail: The level of detail to return - `basic` or `extended`
        :type detail: str.
        :param account_id: Display all system keys belonging to `account_id`
        :type account_id: int.
        :param user_id: Display all keys belonging to `user_id`
        :type user_id: int.
        :returns: `list` - of :class:`ApiKey` or :attr:`access_key`
        """

        if 'access_key' in kwargs:
            r = Resource(cls.PATH + "/" + kwargs['access_key'], endpoint=endpoint)
            params = {}
        else:
            r = Resource(cls.PATH, endpoint=endpoint)

            if 'detail' in kwargs:
                r.request_details = kwargs['detail']
            else:
                r.request_details = 'basic'

            if 'account_id' in kwargs:
                params = {'accountId': kwargs['account_id']}
            elif 'user_id' in kwargs:
                params = {'userId': kwargs['user_id']}
            else:
                params = {}

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)]
                        for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i)
                        for i in uncamel_keys(x)[uncamel(cls.COLLECTION_NAME)]]
        else:
            raise ApiKeyException(r.last_error)


class ApiKeyException(BaseException):
    pass


class ApiKeyGenerationException(ApiKeyException):
    pass


class ApiKeyInvalidationException(ApiKeyException):
    pass



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys, uncamel_keys


class Personality(Resource):

    """ A Dell Cloud Manager personality represents an ordered collection of scripts and/or
    personalities that are executed in an environment to bring a server to fit a certain role. """
    PATH = 'automation/Personality'
    COLLECTION_NAME = 'personalities'
    PRIMARY_KEY = 'cmAccountId'

    def __init__(self, cmAccountId=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__cmAccountId = cmAccountId

    @property
    def cmAccountId(self):
        return self.__cmAccountId

    @classmethod
    def all(cls, cmAccountId, endpoint=None, **kwargs):
        r = Resource(cls.PATH, endpoint=endpoint)

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        params = {'cmAccountId': cmAccountId}
        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise PersonalityException(r.last_error)


class PersonalityException(BaseException):
    pass



"""
mixcoatl.resource
------------------
"""
import os
import requests as r
import mixcoatl.auth as auth
from mixcoatl.settings.load_settings import settings
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camel_keys
import json
import datetime


class Endpoint(object):

    """ A class for representing DCM endpoints and loading them from json files. Each Endpoint contains not only the URL
    for the endpoint but also the associated API keys.

    :param url: The url of the DCM endpoint (e.g https://dcm.enstratius.com/api/enstratus/2015-05-25 )
    :param basepath: Optional server path to the API (e.g. /api/enstratus/2015-05-25)
    :param api_version: required API version (e.g. 2015-05-25) , if missing an attempt is made to parse it
                        from from the end of the url
    :param access_key: DCM access key
    :param secret_key: DCM secret key
    :param ssl_verify: Should the endpoints SSL key be verified. Default is True
    :param nickname: optional arbitrary short nick name to assign this endpoint

    Here's an example of how you could use endpoints to identify servers simultaneously managed by two DCM instances:

        .. code-block:: python

            from mixcoatl.infrastructure.server import Server
            from mixcoatl.resource import Endpoint

            saas_endpoint = Endpoint(nickname="saas",
                                     url="https://dcm.enstratius.com/api/enstratus/2015-05-25",
                                     api_version="2015-05-25",
                                     access_key="POIUYTREQ",
                                     secret_key="ukjdf8HydvkLlki=4hfksj")

            vagrant_endpoint = Endpoint(nickname="vagrant",
                                        url="https://vagrant.vm/api/enstratus/2015-05-25",
                                        api_version="2015-05-25",
                                        access_key="LKJHGFDSAQ",
                                        secret_key="ht748fbsd974d=t874gDFb",
                                        ssl_verify=False)


            saas_servers = Server.all(endpoint=saas_endpoint)
            vagrant_servers = Server.all(endpoint=vagrant_endpoint)

            # find all the servers managed by two DCM instances
            for s in saas_servers:
                for v in vagrant_servers:
                    if s.provider_id == v.provider_id:
                        print "SaaS: %s Vagrant: %s, Provider ID %s" % (s.server_id, v.server_id, v.provider_id)




    """

    def __init__(self, url=None, basepath=None, api_version=None, secret_key=None,
                 access_key=None, ssl_verify=True, nickname=None):
        """

        """
        if url:
            self.url = url
        else:
            raise ValueError("endpoint url must be specified")
        if api_version:
            self.api_version = self._validate_api_version(api_version)
        else:
            try:
                # attempt to grab the API version from the url
                self.api_version = self._validate_api_version(self.url.split('/')[-1])
            except ValueError:
                raise ValueError("endpoint was created without an api_version and did not contain a " +
                                 "URL terminated with a valid date in YYYY-MM-DD format")
        if basepath:
            self.basepath = basepath
        else:
            self.basepath = "/api/enstratus/%s" % self.api_version
        if access_key:
            self.access_key = access_key
        else:
            raise ValueError("access_key must be specified")
        if secret_key:
            self.secret_key = secret_key
        else:
            raise ValueError("secret_key must be specified")

        self.ssl_verify = ssl_verify
        self.nickname = nickname

        if self.ssl_verify is False:
            try:
                r.packages.urllib3.disable_warnings()
            except AttributeError:
                # requests is probably old and doesn't do warnings
                pass
        self.available_api_versions = None
        self.loaded = False

    @staticmethod
    def _validate_api_version(date):
        try:
            datetime.datetime.strptime(date, '%Y-%m-%d')
            return date
        except ValueError:
            raise ValueError("Incorrectly formatted API version, should be YYYY-MM-DD")

    @classmethod
    def from_file(cls, filename):
        """ Load a DCM endpoint definition from a json file and return and Endpoint object.

        :param str filename: the full path to a json file containing and endpoint definition
        :returns: the endpoint loaded from the file
        :rtype: Endpoint

        An example json file containing the containing the endpoint definition:

            .. code-block:: json
            
                {
                  "nickname":"saas",
                  "url":"https://dcm.enstratius.com/api/enstratus/2015-05-25",
                  "api_version":"2015-05-25",
                  "access_key":"abcdefg",
                  "secret_key":"gfedcba",
                  "ssl_verify": true
                }
        """

        endpoint_json = open(filename).read()
        e = json.loads(endpoint_json)

        return Endpoint(url=e['url'],
                        basepath=e['basepath'] if 'basepath' in e else None,
                        api_version=e['api_version'],
                        secret_key=str(e['secret_key']),
                        access_key=str(e['access_key']),
                        ssl_verify=e['ssl_verify'] if 'ssl_verify' in e else None,
                        nickname=e['nickname'] if 'nickname' in e else None)

    @classmethod
    def multiple_from_file(cls, filename):
        """ Load multiple DCM endpoint definitions from a json file and return a dictionary indexed by Endpoint
         nicknames.

        :param str filename: the full path to a json file containing and multiple endpoint definitions
        :returns: a dictionary of Endpoints indexed by endpoint nicknames
        :rtype: dictionary of Endpoint objects

        Here is an example json file that can be loaded by this function:

            .. code-block:: json

                [
                  {
                    "nickname": "saas",
                    "url": "https://dcm.enstratius.com/api/enstratus/2015-05-25",
                    "api_version": "2015-05-25",
                    "access_key": "abcdefg",
                    "secret_key": "gfedcba",
                    "ssl_verify": true
                  },
                  {
                    "nickname": "vagrant",
                    "url": "https://vagrant.vm/api/enstratus/2015-05-25",
                    "api_version": "2015-05-25",
                    "access_key": "abcdefg",
                    "secret_key": "gfedcba",
                    "ssl_verify": true
                  }
                ]
        """

        endpoint_json = open(filename).read()
        endpoint_in = json.loads(endpoint_json)
        endpoint_dict = {}

        for e in endpoint_in:
            endpoint_dict[e['nickname']] = Endpoint(url=e['url'],
                                                    basepath=e['basepath'] if 'basepath' in e else None,
                                                    api_version=e['api_version'],
                                                    secret_key=str(e['secret_key']),
                                                    access_key=str(e['access_key']),
                                                    ssl_verify=e['ssl_verify'] if 'ssl_verify' in e else None,
                                                    nickname=e['nickname'] if 'nickname' in e else None)
        return endpoint_dict

    def _load_api_versions(self):
        from mixcoatl.utils import uncamel_keys

        headers = {'content-type': 'application/json', 'Accept': 'application/json'}
        url_parts = self.url.split("/")
        url = url_parts[0]+"//"+url_parts[2]+"/api/enstratus"
        result = r.get(url,headers=headers)
        result.raise_for_status()
        self.loaded = True
        self.available_api_versions = uncamel_keys(result.json())

    def load(self):
        """Force reload of endpoint attributes such as available API versions remote endpoint via API calls"""
        self._load_api_versions()

    def latest_api_version(self):
        """Get the lastest API version from the DCM enpoint. Note this is could be different from the currently
        active DCM endpoint available at Endpoint.api_version."""

        if self.loaded is False:
            self._load_api_versions()
        return self.available_api_versions["current"]["version_id"]

    def supported_api_versions(self):
        """Get the supported API versions in from the DCM endpoint"""

        if self.loaded is False:
            self._load_api_versions()
        return sorted([ v["version_id"] for v in self.available_api_versions["supported"] ], reverse=True)

class Resource(object):

    """The base class for all resources returned from an DCM API call
    By default all resources are largely represented as a `dict`-alike object
    that mirrors the JSON response from the DCM API with keys converted
    from camel-case to snake-case.

    For instance:

        * The DCM API JSON for a resource looks something like this:

            .. code-block:: yaml

                {
                  'someId':1,
                  'name':'my-name',
                  'cloud':{'cloudId':12345}
                }

        * All keys (nested or top-level) will be converted to snake-case:
            - `someId` becomes `some_id`
            - `cloudId` becomes `cloud_id`
        * The top-level keys will be converted to getters (and in some cases - setters) on the Resource.
            - `some_id` will now be the resource's primary identifier: `resource.some_id`
            - `name` will likely be a getter and a setter as `name` is usually a mutable attributes
            - `cloud` will be converted to a setter with a value of `{'cloud_id':123345}`
        * In cases where a setter would actually need to set a nested value, it will do so. However
            the getter will still return the original data structure as appropriate

        .. warning::

            The Resource object, while looking much like a `dict`, is not a `dict` proper.
            If you need an actual `dict`, call `to_dict()` on your instance.
    """

    #: The base request path of the resource
    PATH = None
    #: The top-level grouping of a resource in the API response
    COLLECTION_NAME = None
    #: The unique identifier of an individual resource
    PRIMARY_KEY = None

    def __init__(self, base_path=None, request_details='basic', endpoint=None, **kwargs):
        if base_path is None:
            try:
                self.__path = self.__class__.PATH
            except:
                raise AttributeError('you must override base_path')
        else:
            self.__path = base_path

        if 'request_details' in kwargs:
            self.__request_details = kwargs['request_details']
            del kwargs['request_details']
        else:
            self.__request_details = 'basic'

        if 'params' in kwargs:
            self.__params = kwargs['params']
        else:
            self.__params = {}
        if endpoint:
            self.__endpoint = endpoint
        else:
            # no endpoint provided, pull from settings
            self.__endpoint = Endpoint(url=settings.endpoint,
                                       basepath=settings.basepath,
                                       api_version=settings.api_version,
                                       secret_key=settings.secret_key,
                                       access_key=settings.access_key,
                                       ssl_verify=settings.ssl_verify)
        self.__last_request = None
        self.__last_error = None
        self.__current_job = None
        self.__status_code = None
        self.__payload_format = 'json'
        self.__request_details = request_details
        self.pending_changes = {}

    def __props(self):
        """List of properties and lazy properties for a given resource"""
        p = [k for k, v in self.__class__.__dict__.items() if type(
            v) in [lazy_property, property]]
        rp = ['last_error', 'path', 'last_request',
              'current_job', 'request_details']
        return p + rp

    def __repr__(self):
        from mixcoatl.utils import convert
        d = {}
        for x in self.__props():
            try:
                if x == 'last_request':
                    d[x] = str(getattr(self, x))
                else:
                    d[x] = getattr(self, x)
            except AttributeError:
                d[x] = None
            except KeyError:
                d[x] = None
        return repr(convert(d))

    @property
    def endpoint(self):
        return self.__endpoint

    @property
    def status_code(self):
        return self.__status_code

    @property
    def request_details(self):
        """The level of detail used in the API call: `basic` or `extended`"""
        return self.__request_details

    @request_details.setter
    def request_details(self, level):
        self.__request_details = level

    @property
    def payload_format(self):
        """The format of the payload: `json` or `xml`"""
        return self.__payload_format

    @payload_format.setter
    def payload_format(self, p_format):
        self.__payload_format = p_format

    @property
    def path(self):
        """The path used in the API request (e.g. ``admin/Job``)"""
        return self.__path

    @path.setter
    def path(self, p):
        self.__path = p

    @property
    def last_request(self):
        """The :class:`Request` object of the most recent API call"""
        return self.__last_request

    @last_request.setter
    def last_request(self, lr):
        self.__last_request = lr

    @property
    def last_error(self):
        """The last error message, if any, returned from the most recent API call"""
        return self.__last_error

    @last_error.setter
    def last_error(self, le):
        self.__last_error = le

    @property
    def current_job(self):
        """The current :class:`Job`, if any, of an asynchronous API call"""
        return self.__current_job

    @current_job.setter
    def current_job(self, cj):
        self.__current_job = cj

    @property
    def params(self):
        return self.__params

    @params.setter
    def params(self, p):
        self.__params = p

    def add_property(self, prop_name):
        """Adds setter, getter and property for field `prop_name`"""
        def set_attribute(cls, val):
            # since we are using __ mangling we need _'+cls.__class__.__name__+
            setattr(cls, '_'+cls.__class__.__name__+'__'+prop_name, val)

        def get_attribute(cls):
            return getattr(cls, '_'+cls.__class__.__name__+'__'+prop_name)

        setattr(self.__class__, 'set_'+prop_name, set_attribute)
        setattr(self.__class__, 'get_'+prop_name, get_attribute)
        setattr(self.__class__,
                prop_name,
                property(
                    getattr(self.__class__, 'get_'+prop_name),
                    getattr(self.__class__, 'set_'+prop_name)
                ))

    def load(self, **kwargs):
        """(Re)load the current object's attributes from an API call"""
        from mixcoatl.utils import uncamel_keys
        reserved_words = ['type']
        p = self.PATH + "/" + str(getattr(self, self.__class__.PRIMARY_KEY))

        if 'params' in kwargs:
            params = kwargs['params']
        else:
            params = camel_keys(self.params)

        s = self.get(p, params=params)
        if self.last_error is None:
            scope = uncamel_keys(s[self.__class__.COLLECTION_NAME][0])
            for k in scope.keys():
                if k in reserved_words:
                    the_key = 'e_' + k
                else:
                    the_key = k
                nk = '_%s__%s' % (self.__class__.__name__, the_key)
                if the_key not in self.__props():
                    self.add_property(k)
                    if 'DCM_DEBUG' in os.environ:
                        print "Missing key found when loading class %s with primary key %s: %s" % \
                              ( self.__class__.__name__, str(getattr(self, self.__class__.PRIMARY_KEY)), k)
                setattr(self, nk, scope[k])
                self.loaded = True
        else:
            return self.last_error

    def __doreq(self, method, **kwargs):
        """Performs the actual API call

        * calls `auth.get_sig` for signed headers
        * issues the requested :attr:`method` against the API endpoint
        * Handles requests appropriately based on sync/async nature of the call
            based on DCM API documentation
        """
        failures = [400, 403, 404, 409, 500, 501, 503]
        sig = auth.get_sig(method, self.path, endpoint=self.__endpoint)
        url = self.__endpoint.url + '/' + self.path
        ssl_verify = self.__endpoint.ssl_verify

        if self.payload_format == 'xml':
            payload_format = 'application/xml'
        elif self.payload_format == 'json':
            payload_format = 'application/json'
        else:
            raise AttributeError(
                'Wrong payload format: %s' % self.payload_format)

        headers = {'x-esauth-access': sig['access_key'],
                   'x-esauth-timestamp': str(sig['timestamp']),
                   'x-esauth-signature': str(sig['signature']),
                   'x-es-details': self.request_details,
                   'Accept': payload_format,
                   'User-Agent': sig['ua']}

        results = r.request(
            method, url, headers=headers, verify=ssl_verify, **kwargs)

        if 'DCM_DEBUG' in os.environ:
            print "URL: %s" % (url)
            print "method: %s" % (method)
            for key, value in kwargs.iteritems():
                print "%s = %s" % (key, value)
            for key, value in results.headers.iteritems():
                print "headers: %s = %s" % (key, value)

        self.last_error = None
        self.last_request = results
        self.__status_code = results.status_code

        if self.payload_format == 'xml':
            return results

        if results.status_code in failures:
            try:
                err = results.json()
                self.last_error = err['error']['message']
            except ValueError:
                self.last_error = results.content
            return self.last_error

        if method == 'GET':
            try:
                results.raise_for_status()
                self.last_error = None
                return results.json()
            except r.exceptions.HTTPError:
                try:
                    err = results.json()
                    self.last_error = err['error']['message']
                except ValueError:
                    self.last_error = results.content
                return False
        if method == 'DELETE':
            if results.status_code == 202:
                self.current_job = results.json()['jobs'][0]['jobId']
                return results.json()
            elif results.status_code != 204:
                try:
                    err = results.json()
                    self.last_error = err['error']['message']
                except ValueError:
                    self.last_error = results.content
                return False
            else:
                return True
        if method == 'PUT':
            if results.status_code == 202:
                self.current_job = results.json()['jobs'][0]['jobId']
                return results.json()
            elif results.status_code == 204:
                return True
            else:
                try:
                    err = results.json()
                    self.last_error = err['error']['message']
                except ValueError:
                    self.last_error = results.content
                return False
        if method == 'POST':
            if results.status_code in [201, 202]:
                if results.status_code == 202:
                    self.current_job = results.json()['jobs'][0]['jobId']
                return results.json()
            else:
                try:
                    err = results.json()
                    self.last_error = err['error']['message']
                except ValueError:
                    self.last_error = results.content
                return False
        if method == 'HEAD':
            if results.status_code in [200]:
                try:
                    err = results.json()
                    self.last_error = err['error']['message']
                except ValueError:
                    self.last_error = results.content
                return False

    def set_path(self, path=None):
        if path is not None:
            self.path = path

    def get(self, path=None, **kwargs):
        """Perform an HTTP `GET` against the API endpoint for the current resource"""
        self.set_path(path)
        return self.__doreq('GET', **kwargs)

    def post(self, path=None, **kwargs):
        """Perform an HTTP `POST` against the API endpoint for the current resource"""
        self.set_path(path)
        return self.__doreq('POST', **kwargs)

    def put(self, path=None, **kwargs):
        """Perform an HTTP `PUT` against the API endpoint for the current resource"""
        self.set_path(path)
        return self.__doreq('PUT', **kwargs)

    def head(self, path=None, **kwargs):
        """Perform an HTTP `HEAD` against the API endpoint for the current resource"""
        self.set_path(path)
        return self.__doreq('HEAD', **kwargs)

    def delete(self, path=None, **kwargs):
        """Perform an HTTP `DELETE` against the API endpoint for the current resource"""
        self.set_path(path)
        return self.__doreq('DELETE', **kwargs)

    def pprint(self):
        """The prettyprint formatted representation of the current resource"""
        import pprint
        pprint.pprint(eval(repr(self)))

    def to_dict(self):
        """The `dict` representation of the current resource"""
        return eval(repr(self))

    def track_change(self, var, prev, new):
        if prev == new:
            pass
        else:
            self.pending_changes[var] = {'old': prev, 'new': new}



"""Implements the DCM ServerProduct API"""
from mixcoatl.utils import uncamel, camelize, camel_keys, uncamel_keys
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property


class ServerProduct(Resource):

    """The Server Agent resource provides methods by which you can query the output of the Dell
    Cloud Manager agent and retrieve its log output. In the future this will be expanded such that
    you can also query for basic health information of the server on which the agent is installed. """
    PATH = 'infrastructure/ServerProduct'
    COLLECTION_NAME = 'serverProducts'
    PRIMARY_KEY = 'product_id'

    def __init__(self, product_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__product_id = product_id

    @property
    def product_id(self):
        """`int` - The unique DCM id for this product"""
        return self.__product_id

    @lazy_property
    def cloud(self):
        """`dict` - The cloud in which this server product is valid"""
        return self.__cloud

    @lazy_property
    def architecture(self):
        """`str` - The CPU architecture of the underlying vCPUs"""
        return self.__architecture

    @lazy_property
    def cpu_count(self):
        """`int` - The number of vCPUs for this product"""
        return self.__cpu_count

    @lazy_property
    def cpu_speed_in_mhz(self):
        """`int` - The speed of the vCPUs for this product"""
        return self.__cpu_speed_in_mhz

    @lazy_property
    def currency(self):
        """`str` - The currenct in which the associated costs are charged"""
        return self.__currency

    @lazy_property
    def description(self):
        """`str` - A detailed description of what this product represents"""
        return self.__description

    @lazy_property
    def disk_size_in_gb(self):
        """`int` - The size of the root volume in GB"""
        return self.__disk_size_in_gb

    @lazy_property
    def hourly_rate(self):
        """`int` - The hourly rate for usage of this product"""
        return self.__hourly_rate

    @lazy_property
    def name(self):
        """`str` - The user-friendly name of this product"""
        return self.__name

    @lazy_property
    def platform(self):
        """`str` - The operating system for this product"""
        return self.__platform

    @lazy_property
    def provider_product_id(self):
        """`str` - The cloud provider's unique id for this product"""
        return self.__provider_product_id

    @lazy_property
    def provider_region_id(self):
        """`str` - The cloud provider region id where this product is valid"""
        return self.__provider_region_id

    @lazy_property
    def ram_in_mb(self):
        """`int` - The memory size for this product in MB"""
        return self.__ram_in_mb

    @lazy_property
    def software(self):
        """`str` - Any additional software installed on this product"""
        return self.__software

    @classmethod
    def all(cls, region_id, endpoint=None, **kwargs):
        """Return all server products

        :param region_id: The region id to search in
        :type region_id: int.
        :param keys_only: Return :attr:`product_id` or :class:`ServerProduct`
        :type keys_only: bool.
        :param detail: Level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` of :attr:`product_id` or :class:`ServerProduct`
        :raises: :class:`ServerProductException`
        """
        r = Resource(cls.PATH, endpoint=endpoint)
        r.request_details = 'basic'
        params = {'regionId': region_id}
        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[uncamel(cls.COLLECTION_NAME)]]
        else:
            raise ServerProductException(r.last_error)


class ServerProductException(BaseException):
    pass



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import uncamel, camelize, camel_keys, uncamel_keys
import json


class ConfigurationManagementService(Resource):

    """ A configuration management service is an actual service endpoint running a supported
    configuration management system. Some services like the public OpsCode Platform for Chef
    may be available to all customers, or you may define your own configuration management
    service for your companys private use. """
    PATH = 'automation/ConfigurationManagementService'
    COLLECTION_NAME = 'cmServices'
    PRIMARY_KEY = 'cm_system_id'

    def __init__(self, cm_account_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)

    @lazy_property
    def cm_system(self):
        return self.__cm_system

    @lazy_property
    def service_endpoint(self):
        return self.__service_endpoint

    @lazy_property
    def properties(self):
        return self.__properties

    @lazy_property
    def cm_service_id(self):
        return self.__cm_service_id

    @lazy_property
    def budget(self):
        return self.__budget

    @budget.setter
    def budget(self, b):
        self.__budget = b

    @lazy_property
    def removable(self):
        return self.__removable

    @lazy_property
    def description(self):
        return self.__description

    @description.setter
    def description(self, b):
        self.__description = b

    @lazy_property
    def name(self):
        return self.__name

    @name.setter
    def name(self, b):
        self.__name = b

    @lazy_property
    def status(self):
        return self.__status

    @lazy_property
    def customer(self):
        return self.__customer

    @required_attrs(['budget', 'description', 'name', 'endpoint', 'cm_system_id'])
    def create(self):
        """Creates a new CM service."""
        parms = [{"budget": self.budget,
                  "serviceEndpoint": self.endpoint,
                  "description": self.description,
                  "name": self.name,
                  "label": "red",
                  "cmSystem": {"cmSystemID": self.cm_system_id}}]

        payload = {"addService": camel_keys(parms)}
        print json.dumps(payload)

        response = self.post(data=json.dumps(payload))
        if self.last_error is None:
            self.load()
            return response
        else:
            raise CMCreationException(self.last_error)

    @classmethod
    def all(cls, endpoint=None, **kwargs):
        r = Resource(cls.PATH, endpoint=endpoint)

        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        x = r.get()
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[uncamel(cls.COLLECTION_NAME)]]
        else:
            raise CMException(r.last_error)


class CMException(BaseException):
    pass


class CMCreationException(CMException):

    """CM Creation Exception"""
    pass



import json
import xmltodict
"""Common helper utilities for use with mixcoatl"""

def to_csv(data):
    """Formats a list of dictionaries into a CSV

    :param data: a list of dictionaries
    :returns str: a str formatted in csv

    .. warning::

        A python dictionary object can't be fully represented in in csv form. Hence, the result returned will
        contain cells that have a text representation of a sub elements.
    """
    csv = ""
    for i in data[0].keys():
        csv += '"'+i.replace('"','\"')+'"'+","
    csv = csv.rstrip(",")

    for x in data:
        csv += "\n"
        for i in x:
            if isinstance(x[i], dict):
                if x[i]:
                    csv += '"'+str(x[i][x[i].keys()[0]]).replace('"','\"')+'"'+","
                else:
                    csv += '"",'
            else:
                csv += '"'+str(x[i]).replace('"','\"')+'"'+","
        csv = csv.rstrip(",")

    return csv

def print_format(data, payload_format):
    """Format the public instance variables of list of objects into either json, xml or csv . The contents of `__dict__`
    including those instance variables that are name mangled using __

    :param data: a list object to be formatted
    :param str payload_format: the the format that should be returned, either "json","xml","csv"
    :returns str: formatted string
    """


    newdata = []
    for i in data:
        thedict = {}
        for key, value in i.__dict__.iteritems():
            if '__' not in key:
                thedict[key] = value
        newdata.append(thedict)

    if payload_format == "xml":
        return "%s %s %s"% ('<?xml version="1.0" ?><root>',
                            xmltodict.unparse({"item":newdata}, full_document=False),
                            '</root>')
    elif payload_format == "csv":
        return to_csv(newdata)
    else:
        return json.dumps(newdata, indent=4, sort_keys=True)


def format_dict(thedict, payload_format):
    """Format a dictionary into json, xml, or csv and return a string

    :param dict thedict: dictionary to be formatted
    :param str payload_format: the format to return either "json", "xml", or "csv"
    :returns str: formatted string
    """


    if payload_format == "json":
        return json.dumps(thedict, indent=4, sort_keys=True)
    elif payload_format == "xml":
        return "%s %s %s"% ('<?xml version="1.0" ?><root>',
                            xmltodict.unparse({"item":thedict}, full_document=False),
                            '</root>')
    elif payload_format == "csv":
        return to_csv([thedict])
    else:
        raise ValueError("Unknown format requested, use payload_format json, xml, csv")


def add_argparse_output_formats(parser):
    """Add the supported cli output formats to the :attr:`parser` ArgumentParser object. Flags added are:
    `--json --xml --csv`

    :param ArgumentParser parser: object parser to append flags to"""

    parser.add_argument('--json', dest='format', action='store_const', const='json',
                        help='print API response in JSON format.')
    parser.add_argument('--xml', dest='format', action='store_const', const='xml',
                    help='print API response in XML format.')
    parser.add_argument('--csv', dest='format', action='store_const', const='csv',
                    help='print API response in CSV format.')


def uncamel(val):
    """Return the snake case version of :attr:`str`

    >>> uncamel('deviceId')
    'device_id'
    >>> uncamel('dataCenterName')
    'data_center_name'
    """
    import re
    s = lambda val: re.sub('(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', '_\\1', val).lower().strip('_')
    return s(val)

def uncamel_keys(d1):
    """Return :attr:`d1` with all keys converted to snake case

    >>> d = {'myThings':[{'thingId':1,'someThings':{'firstThing':'a_thing'}}]}
    >>> uncamel_keys(d)
    {'my_things': [{'thing_id': 1, 'some_things': {'first_thing': 'a_thing'}}]}
    """
    d2 = dict()
    if not isinstance(d1, dict):
        return d1
    for k, v in d1.iteritems():
        new_key = uncamel(k)
        if isinstance(v, dict):
            d2[new_key] = uncamel_keys(v)
        elif isinstance(v, list):
            d2[new_key] = [uncamel_keys(item) for item in v]
        else:
            d2[new_key] = v
    return d2


def camelize(val):
    """Return the camel case version of a :attr:`str`

    >>> camelize('this_is_a_thing')
    'thisIsAThing'
    """
    s = ''.join([t.title() for t in val.split('_')])
    return s[0].lower()+s[1:]


def camel_keys(d1):
    """Return :attr:`d1` with all keys converted to camel case

    >>> b = {'my_things': [{'thing_id': 1, 'some_things': {'first_thing': 'a_thing'}}]}
    >>> camel_keys(b)
    {'myThings': [{'thingId': 1, 'someThings': {'firstThing': 'a_thing'}}]}
    """
    d2 = dict()
    if not isinstance(d1, dict):
        return d1
    for k, v in d1.iteritems():
        new_key = camelize(k)
        if isinstance(v, dict):
            d2[new_key] = camel_keys(v)
        elif isinstance(v, list):
            d2[new_key] = [camel_keys(item) for item in v]
        else:
            d2[new_key] = v
    return d2


def convert(val):
    """Return :attr:`input` converted from :class:`unicode` to :class:`str`

    >>> convert(u'bob')
    'bob'
    >>> convert([u'foo', u'bar'])
    ['foo', 'bar']
    >>> convert({u'foo':u'bar'})
    {'foo': 'bar'}
    """
    if isinstance(val, dict):
        return dict((convert(key), convert(value)) for key, value in val.iteritems())
    elif isinstance(val, list):
        return [convert(element) for element in val]
    elif isinstance(val, unicode):
        return val.encode('utf-8')
    else:
        return val



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import uncamel, camelize, camel_keys, uncamel_keys
import json


class ConfigurationManagementSystem(Resource):

    """A configuration management system represents a product supported by Dell Cloud Manager for
    configuration management. It might be a commercial product like Chef or Puppet or a custom,
    in-house system. Configuration management systems are configured by a Dell Cloud Manager
    administrator. In the SaaS environment, you may request through the help desk system for Dell
    Cloud Manager to include your configuration management system of choice if it isn’t one that
    is supported. """
    PATH = 'automation/ConfigurationManagementSystem'
    COLLECTION_NAME = 'cmSystems'
    PRIMARY_KEY = 'cm_system_id'

    def __init__(self, cm_account_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)

    @classmethod
    def all(cls, endpoint=None, **kwargs):
        r = Resource(cls.PATH, endpoint=endpoint)

        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        x = r.get()
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[uncamel(cls.COLLECTION_NAME)]]
        else:
            raise CMException(r.last_error)


class CMException(BaseException):
    pass


class CMCreationException(CMException):

    """CM Creation Exception"""
    pass



from mixcoatl.resource import Resource
from mixcoatl.admin.job import Job
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import uncamel, camelize, camel_keys, uncamel_keys


class StorageObject(Resource):
    """A blob or container in which blobs are stored in a cloud storage system."""
    PATH = 'platform/StorageObject'
    COLLECTION_NAME = 'storageObjects'
    PRIMARY_KEY = 'storage_object_id'

    def __init__(self, storage_object_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__storage_object_id = storage_object_id

    @property
    def storage_object_id(self):
        """`int` - The unique ID of this storage object."""
        return self.__storage_object_id

    @lazy_property
    def budget(self):
        """`int` - The budget to which the data stored in a container/bucket is billed against."""
        return self.__budget

    @lazy_property
    def e_type(self):
        """`str` - The type of storage object."""
        return self.__e_type

    @lazy_property
    def cloud(self):
        """`dict` - The cloud for which this storage object rests."""
        return self.__cloud

    @lazy_property
    def container(self):
        """`dict` - The container in which this storage object rests."""
        return self.__container

    @lazy_property
    def find_free_name(self):
        """`bool` - If this field is true and there is a name conflict in the cloud DCM will append numbers to the end of the name until it finds an available name for the storage object. If the field is false and there is a name conflict an error will be thrown."""
        return self.__find_free_name

    @lazy_property
    def created_timestamp(self):
        """`str` - The time when this storage object was initially created in cloud storage."""
        return self.__created_timestamp

    @lazy_property
    def label(self):
        """`str` - A label for displaying the storage object."""
        return self.__label

    @lazy_property
    def last_modified_timestamp(self):
        """`str` - The time when this storage object last had changes made to it."""
        return self.__last_modified_timestamp

    @lazy_property
    def name(self):
        """`str` - A user-friendly name you use to identify the storage object."""
        return self.__name

    @lazy_property
    def owning_group(self):
        """`str` - The group owning this container and governing permissions on the container."""
        return self.__owning_group

    @lazy_property
    def owning_user(self):
        """`str` - The user who owns this storage object."""
        return self.__owning_user

    @lazy_property
    def provider_id(self):
        """`str` - The unique identifier in the cloud storage system for this storage object."""
        return self.__provider_id

    @lazy_property
    def read_any(self):
        """`bool` - Indicates that anyone with access to this account can read items stored in the container. Only for containers."""
        return self.__read_any

    @lazy_property
    def read_code(self):
        """`bool` - Indicates that anyone with provisioning rights to the billing code under which this container is billed can read items stored in the container. Only for containers."""
        return self.__read_code

    @lazy_property
    def read_group(self):
        """`bool` - Indicates that anyone in the same group owning this container can read items stored in the container. Only valid for containers."""
        return self.__read_group

    @lazy_property
    def read_public(self):
        """`bool` - Indicates that items in this container should be made public to the entire world. Only valid for containers."""
        return self.__read_public

    @lazy_property
    def region(self):
        """`str` - The region in which this object is stored."""
        return self.__region

    @lazy_property
    def size(self):
        """`int` - The file size for the storage object. Only valid for objects of type 'blob'."""
        return self.__size

    @lazy_property
    def write_any(self):
        """`bool` - Indicates that anyone with access to this account can write to this container. Valid only for containers."""
        return self.__write_any

    @lazy_property
    def write_code(self):
        """`bool` - Indicates that anyone with provisioning rights to the billing code under which this container is billed can write to the container. Valid only for containers."""
        return self.__write_code

    @lazy_property
    def write_group(self):
        """`bool` - Indicates that anyone in the same group owning this container can write to this container. Valid only for containers."""
        return self.__write_group

    def reload(self):
        """Reload resource data from API calls"""
        if self.storage_object_id is not None:
            self.load()
        elif self.current_job is None:
            self.load()
        else:
            if Job.wait_for(self.current_job):
                job = Job(self.current_job, self.endpoint)
                self.__storage_object_id = job.message
                self.load()
            else:
                return self.last_error

    @classmethod
    def all(cls, region_id, endpoint=None, **kwargs):
        """Get a list of all known storage objects.

        >>> StorageObject.all(region_id=100)
        [{'storage_object_id':1,...},{'storage_object_id':2,...}]

        :returns: list -- a list of :class:`StorageObject`
        :raises: StorageObjectException
        """
        r = Resource(cls.PATH, endpoint=endpoint)
        params = {'regionId': region_id}

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                results = [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                results = [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[uncamel(cls.COLLECTION_NAME)]]
            return results
        else:
            raise StorageObjectException(r.last_error)


class StorageObjectException(BaseException):
    pass



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.admin.job import Job
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
import json


class FirewallRule(Resource):

    """ A firewall rule is an ingress or egress permission that grants or denies the right 
    for traffic from a specific network source to a specific network destination."""
    PATH = 'network/FirewallRule'
    COLLECTION_NAME = 'rules'
    PRIMARY_KEY = 'firewall_rule_id'

    def __init__(self, firewall_rule_id=None, endpoint=None, *args, **kwargs):
        if 'detail' in kwargs:
            self.request_details = kwargs['detail']

        Resource.__init__(self, endpoint=endpoint)
        self.__firewall_rule_id = firewall_rule_id

    @property
    def firewall_rule_id(self):
        """`int` - The unique DCM id for this rule"""
        return self.__firewall_rule_id

    @lazy_property
    def destination(self):
        """`str` - The endpoint for the firewall rule. This could describe a different value 
        depending on the destination type."""
        return self.__destination

    @lazy_property
    def destination_type(self):
        """`enum` - Describes the type of the destination value. Most clouds do not 
        support every destination type."""
        return self.__destination_type

    @lazy_property
    def direction(self):
        """`str` - The direction of the rule: `ingress` or `egress`"""
        return self.__direction

    @direction.setter
    def direction(self, d):
        self.__direction = d

    @lazy_property
    def end_port(self):
        """`int` or `None` - The end port for this rule"""
        return self.__end_port

    @end_port.setter
    def end_port(self, p):
        self.__end_port = p

    @lazy_property
    def firewall(self):
        """`dict` - The firewall to which this rule belongs"""
        return self.__firewall

    @firewall.setter
    def firewall(self, fwid):
        self.__firewall = {'firewall_id': fwid}

    @lazy_property
    def protocol(self):
        """`str` - The network protocol for this rule"""
        return self.__protocol

    @protocol.setter
    def protocol(self, p):
        self.__protocol = p

    @lazy_property
    def source(self):
        """`str` - The starting point for the firewall rule. This could describe a 
        different value depending on the source type."""
        return self.__source

    @lazy_property
    def source_type(self):
        """`enum` - Describes the type of the source value. Most clouds do 
        not support every source type."""
        return self.__source_type

    @lazy_property
    def start_port(self):
        """`int` or `None` - The start port for this rule"""
        return self.__start_port

    @start_port.setter
    def start_port(self, p):
        self.__start_port = p

    @lazy_property
    def rule_provider_id(self):
        """`str` - Provider's firewall rule ID"""
        return self.__rule_provider_id

    @lazy_property
    def permission(self):
        """`enum` - Indicates whether the rule allows or denies traffic."""
        return self.__permission

    @lazy_property
    def precedence(self):
        """`int` - A value indicating the position in the order list by which the firewall should 
        process the rules. Not all clouds support precedence for rule ordering."""
        return self.__precedence

    @required_attrs(['firewall', 'source', 'source_type', 'destination', 'destination_type', 'direction', 'permission', 'protocol'])
    def create(self, **kwargs):
        """Create a new firewall rule

            .. warning::

                Does not currently support adding ICMP rules

        :param reason: Reason for the new rule
        :type reason: str.
        :returns: `bool`
        :raises: :class:`FirewallRuleException`
        """

        if 'reason' not in kwargs:
            reason = 'Added by mixcoatl'
        else:
            reason = kwargs['reason']

        payload = {'add_rule': [{
            'firewall_id': self.firewall['firewall_id'],
            'source': self.source,
            'source_type': self.source_type,
            'destination': self.destination,
            'destination_type': self.destination_type,
            'direction': self.direction,
            'permission': self.permission,
            'protocol': self.protocol,
            'reason': reason}]}

        if self.firewall_rule_id is not None:
            raise FirewallRuleException('Cannot modify existing firewall rule')

        self.post(self.PATH, data=json.dumps(camel_keys(payload)))

        if self.last_error is None:
            return True
        else:
            if self.last_request.status_code == 418:
                return True
            else:
                raise FirewallRuleException(self.last_error)

    @required_attrs(['firewall_rule_id'])
    def remove(self, reason):
        """Remove a firewall rule

        :param reason: Reason for removing the rule
        :type reason: str.
        :returns: `bool`
        :raises: :class:`FirewallRuleException`
        """

        params = {'reason': reason}
        self.delete(
            self.PATH + '/' + str(self.firewall_rule_id), params=params)

        if self.last_error is None:
            return True
        else:
            raise FirewallRuleException(self.last_error)

    @classmethod
    def all(cls, firewall_id, endpoint=None, **kwargs):
        """List all rules for `firewall_id`

        :param firewall_id: The id of the firewall to list rules for
        :type firewall_id: int.
        :param detail: Level of detail to return - `basic` or `extended`
        :type detail: str.
        :param keys_only: Return only :attr:`firewall_rule_id` in results
        :type keys_only: bool.
        :returns: `list` of :attr:`firewall_rule_id` or :class:`FirewallRule`
        :raises: :class:`FirewallRuleException`
        """

        r = Resource(cls.PATH, endpoint=endpoint)
        params = {}

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        params['firewallId'] = firewall_id

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)]
                        for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i)
                        for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise FirewallRuleException(r.last_error)


class FirewallRuleException(BaseException):

    """Generic Exception for FirewallRules"""
    pass



"""
mixcoatl.admin.license
---------------------------

Implements access to the DCM License API
"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import uncamel, camelize, camel_keys, uncamel_keys
import json


class License(Resource):
    """License object encapsulates data about DCM Licensing. This information can only be pulled from the DCM API not
    written to it. Note that this class is different from many of the Resource sub classes in that it does not have a
    PRIMARY_KEY since there can be only one License per DCM instance.
    """

    PATH = 'admin/License'
    COLLECTION_NAME = 'license'

    def __init__(self, endpoint=None):
        Resource.__init__(self, endpoint=endpoint)

    @lazy_property
    def expiration_date_time(self):
        """The DCM expiration date and time in format 2015-06-24T00:49:25.757+0000"""
        return self.__expiration_date_time

    @lazy_property
    def days_until_expiration(self):
        """The Number of days until the DCM license expires"""
        return self.__days_until_expiration

    @lazy_property
    def licensee(self):
        """The DCM licensee"""
        return self.__licensee

    @lazy_property
    def server_limit(self):
        """The number of allowed servers with this license"""
        return self.__server_limit

    @lazy_property
    def valid(self):
        """Is the license presently valid"""
        return self.__valid

    def __props(self):
        """List of the names of DCM license properties"""
        return [k for k, v in self.__class__.__dict__.items() if type(
            v) in [lazy_property, property]]

    def prop_dict(self):
        """A dictionary of the DCM license properties"""
        thedict = {}
        for key in self.__props():
            thedict[key] = getattr(self, key)
        return thedict

    def to_json(self):
        """A json string of the DCM license properties"""
        return json.dumps(self.prop_dict(), indent=4, sort_keys=True)

    def load(self, **kwargs):
        """(Re)load the current object's attributes from an API call"""
        from mixcoatl.utils import uncamel_keys

        reserved_words = ['type']
        p = self.PATH

        if 'params' in kwargs:
            params = kwargs['params']
        else:
            params = camel_keys(self.params)

        s = self.get(p, params=params)
        if self.last_error is None:
            scope = uncamel_keys(s)
            for k in scope.keys():
                if k in reserved_words:
                    the_key = 'e_' + k
                else:
                    the_key = k
                nk = '_%s__%s' % (self.__class__.__name__, the_key)
                # nk = '__%s' %  the_key
                if the_key not in self._Resource__props():
                    self.add_property(k)
                setattr(self, nk, scope[k])
                self.loaded = True
        else:
            return self.last_error



from functools import wraps

def required_attrs(attrs):
    """decorator for marking a list of attributes required before calling a function

    :param attrs: List of attributes that must be set
    :type attrs: list.
    :raises: `ValidationException`
    """
    def wrapper(method):
        @wraps(method)
        def validate(obj, *args, **kwargs):
            for ra in attrs:
                try:
                    if getattr(obj, ra) is None:
                        raise ValidationException('Required attribute "%s" is missing' % ra)
                except AttributeError:
                    raise ValidationException('Required attribute "%s" is missing' % ra)
            return method(obj, *args, **kwargs)
        return validate
    return wrapper

class ValidationException(Exception): pass


"""
mixcoatl.admin.role
--------------------

Implements access to the DCM Role API

"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
import json


class Role(Resource):

    """A role defines a common set of permissions that govern access into a given account"""
    PATH = 'admin/Role'
    COLLECTION_NAME = 'roles'
    PRIMARY_KEY = 'role_id'

    def __init__(self, role_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__role_id = role_id

    @property
    def role_id(self):
        """`int` - The unique id of the role"""
        return self.__role_id

    @lazy_property
    def acl(self):
        """`dict` - The access permissions associated with the role"""
        return self.__acl

    @lazy_property
    def description(self):
        """`str` - A user-friendly description of the role"""
        return self.__description

    @description.setter
    def description(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__description = b

    @lazy_property
    def customer(self):
        """`dict` - The customer to whom this role belongs"""
        return self.__customer

    @lazy_property
    def name(self):
        """`str` - The name of the role"""
        return self.__name

    @name.setter
    def name(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__name = b

    @lazy_property
    def status(self):
        """`str` - The status of the role in DCM"""
        return self.__status

    @required_attrs(['role_id'])
    def grant(self, acls):
        """Updates Roles ACL list."""
        p = '%s/%s' % (self.PATH, str(self.role_id))
        payload = {'grant': [{'acl': camel_keys(acls.values()[0])}]}
        return self.put(p, data=json.dumps(payload))
        if self.last_error is None:
            self.load()
        else:
            raise setACLException(self.last_error)

    @required_attrs(['name', 'description'])
    def create(self):
        """Creates a new role. Status is hard-coded to ACTIVE for now. """

        parms = [{'status': "ACTIVE",
                  'name': self.name,
                  'description': self.description}]

        payload = {'addRole': camel_keys(parms)}

        response = self.post(data=json.dumps(payload))
        if self.last_error is None:
            self.load()
            return response
        else:
            raise RoleCreationException(self.last_error)

    @classmethod
    def all(cls, keys_only=False, endpoint=None, **kwargs):
        """Get all roles

        .. note::

            The keys used to make the request determine results visibility

        :param keys_only: Only return :attr:`role_id` instead of :class:`Group` objects
        :type keys_only: bool.
        :param detail: The level of detail to return - `basic` or `extended`
        :type detail: str.
        :param account_id: List roles with mappings to groups in the specified account
        :type account_id: int.
        :param group_id: Provides the role associated with the specified group
        :type group_id: int.
        :returns: `list` of :attr:`role_id` or :class:`Role`
        :raises: :class:`RoleException`
        """
        r = Resource(cls.PATH, endpoint=endpoint)
        params = {}

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'account_id' in kwargs:
            params['account_id'] = kwargs['account_id']

        if 'group_id' in kwargs:
            params['group_id'] = kwargs['group_id']

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise RoleException(r.last_error)


class RoleException(BaseException):
    pass


class setACLException(RoleException):

    """Role Creation Exception"""
    pass


class RoleCreationException(RoleException):

    """Role Creation Exception"""
    pass



import os
import sys
import glob
import shutil
import datetime
from mixcoatl.exceptions import ConfigException


class Config(object):
    # pylint: disable-msg=E0710

    def __init__(self):
        self.access_key = None
        self.secret_key = None
        self.endpoint = None
        self.user_agent = 'mixcoatl'
        self.api_version = None
        self.basepath = None
        self.default_api_version = '2014-07-30'
        self.ssl_verify = None
        self.mixcoatl_dir = os.path.expanduser('~') + "/.mixcoatl"

    def configure(self):
        if not os.path.exists(self.mixcoatl_dir):
            os.makedirs(self.mixcoatl_dir)

        if 'ES_ACCESS_KEY' not in os.environ and 'DCM_ACCESS_KEY' not in os.environ:
            if not os.path.exists(self.mixcoatl_dir + "/default"):
                try:
                    # Python 2
                    prompt = raw_input
                except NameError:
                    # Python 3
                    prompt = input

                cfg_num = 0
                file_list = glob.glob(self.mixcoatl_dir + "/*.config")

                if len(file_list) > 0:
                    for i in file_list:
                        print ">>> Type", cfg_num, "for", i.replace(self.mixcoatl_dir + "/", "").replace(".config", "")
                        cfg_num += 1

                    cfg_pick = prompt("Which config would you like to use? ")

                    shutil.copy(
                        file_list[int(cfg_pick)], self.mixcoatl_dir + "/default")
                else:
                    results = []
                    print "No config file(s) found.  Let us generate one now:"
                    dcm_host = prompt(
                        "What API endpoint would you like to connect to (Example: http[s]://api.endpoint.domain/api/enstratus/<api version>)?\n")
                    dcm_api_access = prompt("What is the API Access Key?\n")
                    dcm_api_secret = prompt("What is the API Secret Key?\n")
                    dcm_ssl_verify = prompt(
                        "Would you like to disable SSL verification (Y/N)?\n")
                    dcm_config_name = prompt(
                        "What would you like to call this connection (No spaces allowed in name)?\n")

                    if dcm_ssl_verify is ['N', 'n', 'no', 'NO', 'No', 'nO']:
                        dcm_ssl_verify_chk = 0
                    else:
                        dcm_ssl_verify_chk = 1

                    results = ['DCM_ENDPOINT=' + str(dcm_host), 'DCM_ACCESS_KEY=' + str(
                        dcm_api_access), 'DCM_SECRET_KEY=' + str(dcm_api_secret), 'DCM_SSL_VERIFY=' + str(dcm_ssl_verify_chk)]

                    new_file = self.mixcoatl_dir + "/" + \
                        dcm_config_name.replace(' ', '') + ".config"

                    with open(new_file, "w") as out_file:
                        out_file.write('\n'.join(results))

                    shutil.copy(new_file, self.mixcoatl_dir + "/default")

                with open(self.mixcoatl_dir + "/default") as f:
                    for line in f:
                        k, v = line.split('=', 1)
                        os.environ[k] = v.strip()
            else:
                with open(self.mixcoatl_dir + "/default") as f:
                    for line in f:
                        k, v = line.split('=', 1)
                        os.environ[k] = v.strip()

        if self.access_key is None:
            if 'ES_ACCESS_KEY' in os.environ and 'DCM_ACCESS_KEY' not in os.environ:
                os.environ['DCM_ACCESS_KEY'] = os.environ['ES_ACCESS_KEY']

            if 'DCM_ACCESS_KEY' in os.environ:
                self.set_access_key(os.environ['DCM_ACCESS_KEY'])
            else:
                raise ConfigException('missing DCM_ACCESS_KEY')

        if self.secret_key is None:
            if 'ES_SECRET_KEY' in os.environ and 'DCM_SECRET_KEY' not in os.environ:
                os.environ['DCM_SECRET_KEY'] = os.environ['ES_SECRET_KEY']

            if 'DCM_SECRET_KEY' in os.environ:
                self.set_secret_key(os.environ['DCM_SECRET_KEY'])
            else:
                raise ConfigException('missing DCM_SECRET_KEY')

        if self.api_version is None:
            if 'ES_API_VERSION' in os.environ and 'DCM_API_VERSION' not in os.environ:
                os.environ['DCM_API_VERSION'] = os.environ['ES_API_VERSION']

            if 'DCM_API_VERSION' in os.environ:
                self.set_api_version(os.environ['DCM_API_VERSION'])
            else:
                if 'DCM_ENDPOINT' in os.environ:
                    thestring = os.environ['DCM_ENDPOINT'].split('/')

                    if self.validate(thestring[-1]):
                        self.set_api_version(thestring[-1])
                    else:
                        self.set_api_version(self.default_api_version)
                else:
                    self.set_api_version(self.default_api_version)

            self.set_basepath('/api/enstratus/%s' % self.api_version)

        if self.endpoint is None:
            if 'ES_ENDPOINT' in os.environ and 'DCM_ENDPOINT' not in os.environ:
                os.environ['DCM_ENDPOINT'] = os.environ['ES_ENDPOINT']

            if 'DCM_ENDPOINT' in os.environ:
                self.set_endpoint(os.environ['DCM_ENDPOINT'])
            else:
                self.set_endpoint('https://dcm.enstratius.com' + self.basepath)

        if self.ssl_verify is None:
            if 'ES_SSL_VERIFY' in os.environ and 'DCM_SSL_VERIFY' not in os.environ:
                os.environ['DCM_SSL_VERIFY'] = os.environ['ES_SSL_VERIFY']

            if 'DCM_SSL_VERIFY' in os.environ:
                self.set_ssl_verify(os.environ['DCM_SSL_VERIFY'])
            else:
                self.set_ssl_verify('1')

    def validate(self, date_text):
        try:
            datetime.datetime.strptime(date_text, '%Y-%m-%d')
            return date_text
        except ValueError:
            raise ValueError("Incorrect data format, should be YYYY-MM-DD")

    def set_access_key(self, key):
        self.access_key = key

    def set_secret_key(self, key):
        self.secret_key = key

    def set_endpoint(self, ep):
        self.endpoint = ep

    def set_api_version(self, version):
        self.api_version = version

    def set_basepath(self, basepath):
        self.basepath = basepath

    def set_ssl_verify(self, verify):
        if verify == '0':
            self.ssl_verify = False
        else:
            self.ssl_verify = True



"""
mixcoatl.admin.account
----------------------

Implements access to the DCM Account API
"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
import json


class Account(Resource):

    """An account object represents an DCM account held by an DCM customer."""
    PATH = 'admin/Account'
    COLLECTION_NAME = 'accounts'
    PRIMARY_KEY = 'account_id'

    def __init__(self, account_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__account_id = account_id

    @property
    def account_id(self):
        """`int` - The unique ID of this account"""
        return self.__account_id

    @lazy_property
    def alert_configuration(self):
        """`dict` - The configuration of alert preferences for this account."""
        return self.__alert_configuration

    @lazy_property
    def billing_system_id(self):
        """`int` - The ID that may appear on invoices."""
        return self.__billing_system_id

    @lazy_property
    def cloud_subscription(self):
        """`dict` or `None` -- Information about the cloud for this account"""
        return self.__cloud_subscription

    @lazy_property
    def cloud_subscription_id(self):
        """`int` -- Cloud subscription ID """
        return self.__cloud_subscription_id

    @lazy_property
    def configured(self):
        """`bool` - Has account has been tied to an account with a cloud"""
        return self.__configured

    @lazy_property
    def customer(self):
        """`dict` - The DCM customer record to which this account belongs"""
        return self.__customer

    @lazy_property
    def default_budget(self):
        """The billing id that is the default for discovered resources"""
        return self.__default_budget

    @lazy_property
    def dns_automation(self):
        """Does this account subscribe to dns_automation?"""
        return self.__dns_automation

    @lazy_property
    def name(self):
        """`str` - User-friendly name used to identify the account"""
        return self.__name

    @lazy_property
    def owner(self):
        """`dict` - user who owns this account"""
        return self.__owner

    @lazy_property
    def plan_id(self):
        """`int` - pricing plan associated with this account"""
        return self.__plan_id

    @lazy_property
    def provisioned(self):
        """`bool` - Is this account in goodstanding and managed by DCM"""
        return self.__provisioned

    @lazy_property
    def status(self):
        """`str` - The current account payment status"""
        return self.__status

    @lazy_property
    def subscribed(self):
        """`bool` - If the account is configured & working with DCM"""
        return self.__subscribed

    @lazy_property
    def account_name(self):
        return self.__account_name

    @account_name.setter
    def account_name(self, n):
        self.__account_name = n

    @lazy_property
    def account_number(self):
        return self.__account_number

    @account_number.setter
    def account_number(self, n):
        self.__account_number = n

    @lazy_property
    def api_key_id(self):
        return self.__api_key_id

    @api_key_id.setter
    def api_key_id(self, n):
        self.__api_key_id = n

    @lazy_property
    def api_key_secret(self):
        return self.__api_key_secret

    @api_key_secret.setter
    def api_key_secret(self, n):
        self.__api_key_secret = n

    @lazy_property
    def cloud_id(self):
        return self.__cloud_id

    @cloud_id.setter
    def cloud_id(self, n):
        self.__cloud_id = n

    @customer.setter
    def customer(self, n):
        self.__customer = n

    @required_attrs(['account_name', 'account_number', 'cloud_id', 'api_key_id', 'api_key_secret'])
    def add(self):
        payload = {'addAccount': [
            {
                'name': self.account_name,
                'cloudSubscription':
                {
                    'accountNumber': self.account_number,
                    'cloudId': int(self.cloud_id),
                    'apiKeyId': self.api_key_id,
                    'apiKeySecret': self.api_key_secret
                }
            }
        ]}

        if hasattr(self,'customer'):
            if self.customer:
                payload['addAccount'].append({'customer': {'customerId': int(self.customer)}})

        response = self.post(self.PATH, data=json.dumps(payload))
        if self.last_error is None:
            return response
        else:
            raise CreateAccountException(self.last_error)

    @classmethod
    def all(cls, keys_only=False, endpoint=None, **kwargs):
        """Get all accounts

        >>> Account.all(detail='basic')
        [{'account_id':12345,....}]

        >>> Account.all(keys_only=True)
        [12345]

        :param keys_only: Only return `account_id` instead of `Account` objects
        :type keys_only: bool.
        :param detail: The level of detail to return - `basic` or `extended`
        :type detail: str.
        :param cloud_id: Only show accounts tied to the given cloud
        :type cloud_id: int.
        :returns: `list` of :class:`Account` or :attr:`account_id`
        :raises: :class:`AccountException`
        """
        r = Resource(cls.PATH,endpoint=endpoint)
        params = {}

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'cloud_id' in kwargs:
            params = {'cloudId': kwargs['cloud_id']}

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)]
                        for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i)
                        for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise AccountException(r.last_error)

    def assign_cloud(self, cloud_id, account_number, api_key_id, api_key_secret):
        """ Associate this account with the cloud and supporting credentials specified.

        :param cloud_id: Cloud ID of the cloud that will be associated.
        :type cloud_id: int.
        :param account_number: Account number of the cloud credential.
        :type account_number: str.
        :param api_key_id: API access key of the cloud credential.
        :type api_key_id: str.
        :param api_key_secret: API secret key of the cloud credential.
        :type api_key_secret: str.
        :returns: cloud subscription ID.
        :raises: :class:`AssignCloudException`
        """

        p = "%s/%s" % (self.PATH, str(self.account_id))
        payload = {'assignCloud': {
            'accounts': {
                'cloudSubscription': {
                    'cloudId': cloud_id,
                    'accountNumber': account_number,
                    'apiKeyId': api_key_id,
                    'apiKeySecret': api_key_secret}}}}
        self.put(p, data=json.dumps(payload))
        if self.last_error is None:
            self.load()
            return self.cloud_subscription['cloud_subscription_id']
        else:
            raise AssignCloudException(self.last_error)


class AccountException(BaseException):
    pass


class AssignCloudException(AccountException):
    pass


class CreateAccountException(AccountException):
    pass



"""Implements the DCM DataCenter API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import uncamel, camelize, camel_keys, uncamel_keys


class DataCenter(Resource):

    """
    A data center is a part of a regional infrastructure that has some ability
        to share resources with other data centers in the region
    """
    PATH = 'geography/DataCenter'
    COLLECTION_NAME = 'dataCenters'
    PRIMARY_KEY = 'data_center_id'

    def __init__(self, data_center_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__data_center_id = data_center_id

    @property
    def data_center_id(self):
        """`int` - The unique DCM id for this data center"""
        return self.__data_center_id

    @lazy_property
    def description(self):
        """`str` - A description of the data center"""
        return self.__description

    @lazy_property
    def name(self):
        """`str` - A user-friendly name for the data center"""
        return self.__name

    @lazy_property
    def provider_id(self):
        """`str` - The cloud provider's unique id for the data center"""
        return self.__provider_id

    @lazy_property
    def region(self):
        """`dict` = The region to which this data center belongs"""
        return self.__region

    @lazy_property
    def status(self):
        """`str` - The current status of the data center"""
        return self.__status

    @classmethod
    def all(cls, region_id, endpoint=None, **kwargs):
        """Return all data centers

        :param region_id: Required. The region to query against
        :type region_id: int.
        :param keys_only: Return :attr:`data_center_id` instead of :class:`DataCenter`
        :type keys_only: bool.
        :param detail: The level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` of :class:`DataCenter` or :attr:`data_center_id`
        :raises: :class:`DataCenterException`
        """
        r = Resource(cls.PATH, endpoint=endpoint)

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        params = {'regionId': region_id}

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[uncamel(cls.COLLECTION_NAME)]]
        else:
            raise DataCenterException(r.last_error)


class DataCenterException(BaseException):
    pass



import time
import base64
import hashlib
import hmac
from mixcoatl.settings.load_settings import settings


def get_sig(http_method, path, endpoint=None):
    """Return a signature valid for use in making DCM API calls

    :param http_method: The `http_method` used to make the API call
    :param path: The `path` used in the API call
    """

    if endpoint:
        access_key = endpoint.access_key
        secret_key = endpoint.secret_key
        basepath = endpoint.basepath
    else:
        access_key = settings.access_key
        secret_key = settings.secret_key
        basepath = settings.basepath

    timestamp = int(round(time.time() * 1000))
    signpath = basepath + '/' + path
    parts = []
    parts.append(access_key)
    parts.append(http_method)
    parts.append(signpath)
    parts.append(timestamp)
    parts.append(settings.user_agent)
    # pylint: disable-msg=E1101
    dm = hashlib.sha256
    to_sign = ':'.join([str(x) for x in parts])
    d = hmac.new(secret_key, msg=to_sign, digestmod=dm).digest()
    b64auth = base64.b64encode(d).decode()
    return {'timestamp': timestamp,
            'signature': b64auth,
            'access_key': access_key,
            'ua': settings.user_agent}



"""Implements the ability to list CM environments via the API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
from mixcoatl.admin.job import Job


class Environment(Resource):

    """ An Dell Cloud Manager environment represents a Chef environment and is managed in a
    configuration management account.  Environments are NOT uniquely identified by their environmentId. 
    The unique identifier is sharedEnvironmentCode. """
    PATH = 'automation/Environment'
    COLLECTION_NAME = 'environments'
    PRIMARY_KEY = 'environmentId'

    def __init__(self, environmentId=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__environmentId = environmentId

    @classmethod
    def all(cls, cmAccountId, endpoint=None, **kwargs):
        r = Resource(cls.PATH, endpoint=endpoint)

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        params = {'cmAccountId': cmAccountId}
        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise EnvironmentException(r.last_error)


class EnvironmentException(BaseException):
    pass



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import uncamel, camelize, camel_keys, uncamel_keys


class LoadBalancer(Resource):

    """ A load balancer is a virtual load balancer such as an AWS Elastic Load Balancer. It is specifically
    not a VM-hosted load balancer. Load balancers vary wildly from cloud provider to cloud
    provider. As a result, you should check with your cloud meta-data to see what elements are
    necessary in order to create a load balancer. """
    PATH = 'network/LoadBalancer'
    COLLECTION_NAME = 'loadBalancers'
    PRIMARY_KEY = 'load_balancer_id'

    def __init__(self, load_balancer_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__load_balancer_id = load_balancer_id

    @property
    def load_balancer_id(self):
        return self.__load_balancer_id

    @lazy_property
    def customer(self):
        return self.__customer

    @lazy_property
    def status(self):
        return self.__status

    @lazy_property
    def provider_id(self):
        return self.__provider_id

    @lazy_property
    def description(self):
        return self.__description

    @lazy_property
    def region(self):
        return self.__region

    @lazy_property
    def budget(self):
        return self.__budget

    @lazy_property
    def owning_groups(self):
        return self.__owning_groups

    @lazy_property
    def cname_based(self):
        return self.__cname_based

    @lazy_property
    def address(self):
        return self.__address

    @lazy_property
    def owning_account(self):
        return self.__owning_account

    @lazy_property
    def owning_user(self):
        return self.__owning_user

    @lazy_property
    def cloud(self):
        return self.__cloud

    @lazy_property
    def name(self):
        return self.__name

    @lazy_property
    def data_centers(self):
        return self.__data_centers

    @lazy_property
    def servers(self):
        return self.__servers

    @lazy_property
    def listeners(self):
        return self.__listeners

    @classmethod
    def all(cls, endpoint=None, **kwargs):
        r = Resource(cls.PATH, endpoint=endpoint)
        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        x = r.get()
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[uncamel(cls.COLLECTION_NAME)]]
        else:
            raise LoadBalancerException(r.last_error)


class LoadBalancerException(BaseException):
    pass



from mixcoatl.resource import Resource
from mixcoatl.admin.job import Job
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.decorators.lazy import lazy_property
import json
import sys
import time


class Server(Resource):

    """A server is a virtual machine running within a data center."""
    PATH = 'infrastructure/Server'
    COLLECTION_NAME = 'servers'
    PRIMARY_KEY = 'server_id'

    def __init__(self, server_id=None, endpoint=None):
        Resource.__init__(self, endpoint=endpoint)
        self.__server_id = server_id


    @property
    def server_id(self):
        """`int` - The DCM ID of this server"""
        return self.__server_id

    @lazy_property
    def agent_version(self):
        """`int` - The version of the DCM agent if installed."""
        return self.__agent_version

    @lazy_property
    def legacy_owner_id(self):
        return self.__legacy_owner_id

    @lazy_property
    def cloud(self):
        """`dict` - The cloud provided where the instance is located."""
        return self.__cloud

    @lazy_property
    def label(self):
        """`str` - The label assigned to the server"""
        return self.__label

    @label.setter
    def label(self, l):
        self.__label = l

    @lazy_property
    def customer(self):
        """`dict` - The customer account for the server."""
        return self.__customer

    @lazy_property
    def cm_account(self):
        """`dict` - The configuration management account for the server."""
        return self.__cm_account

    @lazy_property
    def data_center(self):
        """`dict` - The specific datacenter where the instance is located."""
        return self.__data_center

    @data_center.setter
    def data_center(self, d):
        self.__data_center = {u'data_center_id': d}

    @lazy_property
    def cmAccount(self):
        return self.__cmAccount

    @lazy_property
    def environment(self):
        return self.__environment

    @environment.setter
    def environment(self, c):
        self.__environment = {u'sharedEnvironmentCode': c}

    @cmAccount.setter
    def cm_account_id(self, c):
        self.__cmAccount = {u'cmAccountId': c}

    @lazy_property
    def cm_scripts(self):
        return self.__cm_scripts

    @lazy_property
    def last_agent_contact_timestamp(self):
        return self.__last_agent_contact_timestamp

    @cm_scripts.setter
    def cm_scripts(self, c):
        s = c.split(",")
        sc = []
        for cm in s:
            sc.append({u'sharedScriptCode': cm})
        self.__cm_scripts = sc

    @lazy_property
    def p_scripts(self):
        return self.__p_scripts

    @p_scripts.setter
    def p_scripts(self, c):
        s = c.split(",")
        p = []
        for cm in s:
            p.append({'sharedPersonalityCode': cm})

        self.__p_scripts = p

    @lazy_property
    def description(self):
        """The description of the server"""
        return self.__description

    @description.setter
    def description(self, d):
        self.__description = d

    @lazy_property
    def machine_image(self):
        """`dict` - The machine image to use/used to provision the server"""
        return self.__machine_image

    @machine_image.setter
    def machine_image(self, m):
        self.__machine_image = {u'machine_image_id': m}

    @lazy_property
    def vlan(self):
        """`list` - The vlan to assign/assigned to the server Deprecated in k.50"""
        return self.__vlan

    @vlan.setter
    def vlan(self, v):
        self.__vlan = {u'vlan_id': v}

    @lazy_property
    def network(self):
        """`list` - The network to assign/assigned to the server"""
        return self.__network

    @network.setter
    def network(self, n):
        self.__network = {u'network_id': n}

    @lazy_property
    def subnet(self):
        """`list` - The subnet to assign/assigned to the server"""
        return self.__subnet

    @subnet.setter
    def subnet(self, n):
        self.__subnet = {u'subnet_id': n}

    @lazy_property
    def firewalls(self):
        """`list` - The firewalls to assign/assigned to the server"""
        return self.__firewalls

    @firewalls.setter
    def firewalls(self, f):
        self.__firewalls = f

    @legacy_owner_id.setter
    def legacy_owner_id(self, f):
        self.__legacy_owner_id = f

    @lazy_property
    def name(self):
        """`str` - The name assigned/to assign to the server"""
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n

    @lazy_property
    def owning_groups(self):
        """`list` - The DCM groups owning the server."""
        return self.__owning_groups

    @lazy_property
    def owning_user(self):
        """`dict` - The DCM user owning the server."""
        return self.__owning_user

    @lazy_property
    def platform(self):
        """`str` - The platform of the server *(i.e. `UBUNTU`)*."""
        return self.__platform

    @lazy_property
    def personalities(self):
        """`dict` - The personalities associated with the server"""
        return self.__personalities

    @lazy_property
    def private_ip_addresses(self):
        """`list` - The private ip addresses assigned to the server."""

        return self.__private_ip_addresses

    @lazy_property
    def public_ip_address(self):
        """`str` - The public ip address of the server."""
        return self.__public_ip_address

    @lazy_property
    def public_ip_addresses(self):
        """`list` - The list of public ip addresses of the server."""
        return self.__public_ip_addresses

    @lazy_property
    def region(self):
        """`dict` - The region where the server is located"""
        return self.__region

    @lazy_property
    def provider_product_id(self):
        """`str` - The provider's product identifier for the server *(i.e. `m1.large`)*"""
        return self.__provider_product_id

    @provider_product_id.setter
    def provider_product_id(self, id):
        self.__provider_product_id = id

    @lazy_property
    def provider_id(self):
        """`str` - The provider's identifier for the server *(i.e. `i-abcdefg`)*"""
        return self.__provider_id

    @lazy_property
    def product(self):
        """`dict` - The DCM product record for the server"""
        return self.__product

    @lazy_property
    def start_date(self):
        """`str` - The date the server was started"""
        return self.__start_date

    @lazy_property
    def stop_date(self):
        """`str` - The date the server was stopped"""
        return self.__stop_date

    @lazy_property
    def status(self):
        """`str` - The status of the server *(i.e. `RUNNING` or `PAUSED`)*."""
        return self.__status

    @lazy_property
    def budget(self):
        """`int` - The budget code applied to the server."""
        return self.__budget

    @budget.setter
    def budget(self, b):
        """`int` - The budget code to apply to the server."""
        self.__budget = b

    @lazy_property
    def scripts(self):
        """`list` - The list of configuration management scripts of the server.(Chef?)"""
        return self.__scripts

    @lazy_property
    def run_list(self):
        """`list` - The list of configuration management scripts of the server.(Puppet?)"""
        return self.__run_list

    @lazy_property
    def architecture(self):
        """`str` - The architecture type of the server."""
        return self.__architecture

    @lazy_property
    def terminate_after(self):
        """`str` - The time the server automatically shuts down."""
        return self.__terminate_after

    @terminate_after.setter
    def terminate_after(self, t):
        self.__terminate_after = t

    @lazy_property
    def pause_after(self):
        """`str` - The time the server automatically pauses."""
        return self.__pause_after

    @lazy_property
    def userData(self):
        """`list` - The userdata to assign/assigned to the server"""
        return self.__userData

    @userData.setter
    def userData(self, userData):
        self.__userData = userData

    @lazy_property
    def volumeConfiguration(self):
        """`type` - The volumeConfiguration to assign/assigned to the server"""
        return self.__volumeConfiguration

    @volumeConfiguration.setter
    def volumeConfiguration(self, volumeConfiguration):
        self.__volumeConfiguration = volumeConfiguration

    @property
    def keypair(self):
        """`str` - The keypair to assign

            .. note::
                DCM does not track keypairs used to launch servers.
                This attribute is used only in the `launch()` call.
        """
        return self.__keypair

    @keypair.setter
    def keypair(self, kp):
        self.__keypair = kp

    def reload(self):
        """Reload resource data from API calls"""
        if self.server_id is not None:
            self.load()
        elif self.current_job is None:
            self.load()
        else:
            if Job.wait_for(self.current_job, endpoint=self.endpoint):
                job = Job(self.current_job, endpoint=self.endpoint)
                self.__server_id = job.message
                self.load()
            else:
                return self.last_error

    @required_attrs(['server_id'])
    def destroy(self, reason='no reason provided'):
        """Terminate server instance with reason :attr:`reason`

        :param reason: The reason for terminating the server
        :type reason: str.
        :returns: bool -- Result of API call
        """
        p = self.PATH + "/" + str(self.server_id)
        qopts = {'reason': reason}
        return self.delete(p, params=qopts)

    @required_attrs(['server_id'])
    def pause(self, reason=None):
        """Pause the server instance with reason :attr:`reason`

        :param reason: The reason for pausing the server
        :type reason: str.
        :returns: Job -- Result of API call
        """
        p = '%s/%s' % (self.PATH, str(self.server_id))
        payload = {'pause': [{}]}

        if reason is not None:
            payload['pause'][0].update({'reason': reason})

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['server_id'])
    def extend_terminate(self, extend):
        p = '%s/%s' % (self.PATH, str(self.server_id))
        qopts = {'terminateAfter': extend}
        return self.delete(p, params=qopts)

    @required_attrs(['server_id'])
    def set_owner(self, owner):
        p = '%s/%s' % (self.PATH, str(self.server_id))
        qopts = {'assignGroups': [{'owningGroups':[{'groupId':owner}]}]}
        #print json.dumps(qopts)
        return self.put(p, data=json.dumps(qopts))

    @required_attrs(['server_id'])
    def start(self, reason=None):
        """Start the paused server instance with reason :attr:`reason`

        :param reason: The reason for starting the server
        :type reason: str.
        :returns: Job -- Result of API call
        """
        p = '%s/%s' % (self.PATH, str(self.server_id))
        payload = {'start': [{}]}
        if reason is not None:
            payload['start'][0].update({'reason': reason})
        return self.put(p, data=json.dumps(payload))

    @required_attrs(['server_id'])
    def provision_user(self, user_id, admin_role=None):
        """Add a user to a server

        :param user_id: The ID of the user
        :type reason: int
        :returns: Job -- Result of API call
        """
        p = '%s/%s' % (self.PATH, str(self.server_id))
        user_dict = {'userId': user_id}
        if admin_role is not None:
            user_dict['adminRole'] = admin_role
        payload = {'provisionUser': [{'user': user_dict}]}
        return self.put(p, data=json.dumps(payload))

    @required_attrs(['server_id'])
    def deprovision_user(self, user_id):
        """Remove a user to a server

        :param user_id: The ID of the user
        :type reason: int
        :returns: Job -- Result of API call
        """
        p = '%s/%s' % (self.PATH, str(self.server_id))
        user_dict = {'userId': user_id}
        payload = {'deprovisionUser': [{'user': user_dict}]}
        return self.put(p, data=json.dumps(payload))

    @required_attrs(['server_id', 'name'])
    def rename(self):
        """Rename server"""
        p = '%s/%s' % (self.PATH, str(self.server_id))
        payload = {'describeServer': [{'name': self.name}]}
        return self.put(p, data=json.dumps(payload))

    @required_attrs(['server_id'])
    def stop(self, reason=None):
        """Stop the server instance with reason :attr:`reason`

        :param reason: The reason for stopping the server
        :type reason: str.
        :returns: Job -- Result of API call
        """
        p = '%s/%s' % (self.PATH, str(self.server_id))
        payload = {'stop': [{}]}

        if reason is not None:
            payload['stop'][0].update({'reason': reason})

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['provider_product_id', 'machine_image', 'description',
                     'name', 'data_center', 'budget'])
    def launch(self, callback=None):
        """Launches a server with the configured parameters

        >>> def cb(j): print(j)
        >>> s = Server()
        >>> s.provider_product_id = 'm1.large'
        >>> s.machine_image = 12345
        >>> s.description = 'my first launch'
        >>> s.name = 'server-1-test'
        >>> s.data_center = 54321
        >>> s.keypair = 'my-aws-keypair'
        >>> s.launch(callback=cb)

        :param callback: Optional callback to send the results of the API call
        :type callback: func.
        :returns: int -- The job id of the launch request
        :raises: :class:`ServerLaunchException`, :class:`mixcoatl.decorators.validations.ValidationException`
        """
        optional_attrs = ['userData', 'vlan', 'subnet', 'network', 'firewalls', 'keypair', 'label',
                          'cmAccount', 'environment', 'cm_scripts', 'p_scripts', 'volumeConfiguration']
        if self.server_id is not None:
            raise ServerLaunchException(
                'Cannot launch an already running server: %s' % self.server_id)

        payload = {'launch':
                   [{
                       'product': {"productId": self.provider_product_id},
                       'budget': self.budget,
                       'machineImage': camel_keys(self.machine_image),
                       'description': self.description,
                       'name': self.name,
                       'dataCenter': camel_keys(self.data_center)
                   }]}

        for oa in optional_attrs:
            try:
                if getattr(self, oa) is not None:
                    if oa == 'cm_scripts':
                        payload['launch'][0].update(
                            {'scripts': getattr(self, oa)})
                    elif oa == 'p_scripts':
                        payload['launch'][0].update(
                            {'personalities': getattr(self, oa)})
                    elif oa == 'volumeConfiguration':
                        payload['launch'][0].update(
                            {'volumeConfiguration': getattr(self, oa)})
                    elif oa == 'vlan':
                        # this was deprecated in k.50
                        payload['launch'][0].update(
                            {'vlan': camel_keys(getattr(self, oa))})
                    elif oa == 'network':
                        payload['launch'][0].update(
                            {'network': camel_keys(getattr(self, oa))})
                    elif oa == 'subnet':
                        payload['launch'][0].update(
                            {'subnet': camel_keys(getattr(self, oa))})
                    else:
                        payload['launch'][0].update({oa: getattr(self, oa)})
            except AttributeError:
                pass

        self.post(data=json.dumps(payload))
        if self.last_error is None:
            if callback is not None:
                callback(self.current_job)
            else:
                return self.current_job
        else:
            raise ServerLaunchException(self.last_error)

    def set_from_file(self, filename):
        """load server attributes from a json file
        :param filename: str with the full path to the json file containing the server attributes
        """
        server_file = open(filename).read()
        server_dict = json.loads(server_file)
        for key,value in server_dict.items():
            setattr(self, key, value)

    def duplicate(self, server):
        pass

    def wait_for(self, status='RUNNING', callback=None):
        """Blocks execution until the current server has status of :attr:`status`

        :param status: The status to expect before continuing *(i.e. `RUNNING` or `PAUSED`)*
        :type status: str.
        :param callback: Optional callback to be called with the final :class:`Server`` when ``status`` is reached
        :type callback: func.
        :raises: `ServerException`
        """
        if self.server_id is None:
            raise ServerException('Must be called with an existing server.')
        initial_status = self.status
        if self.last_error is None:
            if initial_status == status:
                return self
            while self.status != status:
                time.sleep(5)
                self.load()
                if self.last_error is not None:
                    raise ServerException(self.last_error)
                else:
                    continue
            if callback is not None:
                callback(self)
            else:
                return self

    @classmethod
    def all(cls, endpoint=None, **kwargs):
        """Get a list of all known servers

        >>> Server.all()
        [{'server_id':1,...},{'server_id':2,...}]

        :returns: list -- a list of :class:`Server`
        :raises: ServerException
        """
        r = Resource(cls.PATH, endpoint=endpoint)
        params = {}

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        if 'region_id' in kwargs:
            params['regionId'] = kwargs['region_id']

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise ServerException(r.last_error)


class ServerException(BaseException):
    pass


class ServerLaunchException(ServerException):
    pass



"""
mixcoatl.admin.job
------------------
Implements access to the DCM Job API
"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
import time


class Job(Resource):

    """A job is an asynchronous process resulting from a client request that resulted in a 202
    ACCEPTED response. If the client cares about the ultimate result of the original request, it can
    query for the job returned in the initial response until the job completes."""
    PATH = 'admin/Job'
    COLLECTION_NAME = 'jobs'
    PRIMARY_KEY = 'job_id'

    def __init__(self, job_id=None, endpoint=None, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__job_id = job_id

    @property
    def job_id(self):
        """`int` The unique DCM id for this job"""
        return self.__job_id

    @lazy_property
    def status(self):
        """`str` The current status of this job"""
        return self.__status

    @lazy_property
    def description(self):
        """`str` The description of the running job."""
        return self.__description

    @lazy_property
    def message(self):
        """`str` A message describing the current disposition of the operation"""
        return self.__message

    def latest_message(self):
        """`str` The *latest* message describing the current disposition of the operation
            this method forces a reload of the Job attributes from the API. It is often useful
            after waiting for a for a job to complete with wait_for(job_id)
        """
        self.load()
        return self.message

    @lazy_property
    def start_date(self):
        """`str` The date and time when the job was started"""
        return self.__start_date

    @lazy_property
    def end_date(self):
        """`str` The data and time when the job was completed
        `None` if :attr:`status is `RUNNING`
        """
        return self.__end_date

    @lazy_property
    def job_message(self):
        return self.__job_message

    @classmethod
    def all(cls, keys_only=False, endpoint=None):
        """Get all jobs
        :param keys_only: Only return :attr:`job_id` instead of :class:`Job`
        :type keys_only: bool.
        :returns: `list` of :class:`Job` or :attr:`job_id`
        :raises: :class:`JobException`
        """
        r = Resource(cls.PATH, endpoint=endpoint)
        x = r.get()
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)]
                        for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i)
                        for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise JobException(r.last_error)

    @classmethod
    def wait_for(cls, job_id, status='COMPLETE', callback=None, endpoint=None):
        """Blocks execution until :attr:`job_id` returns :attr:`status`

        :param job_id: The ID of the job to wait on
        :type job_id: int.
        :param status: The status to expect before continuing
        :type status: str.
        :param callback: Optional callback to be called with the final :class:`Job`
            when :attr:`status` is reached
        :type callback: func.
        :returns: `bool` - Result of job exectution
        :raises: :class:`JobException`
        """
        j = Job(job_id, endpoint=endpoint)
        j.load()
        if j.last_error is not None:
            raise JobException(j.last_error)
        else:
            while j.status != status:
                time.sleep(5)
                j.load()
                if j.status == 'ERROR':
                    break
                else:
                    continue
        if callback is not None:
            callback(j)
        else:
            return True


class JobException(BaseException):
    pass



#!/usr/bin/env python

import mixcoatl
import os
import glob
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

packages = [
    'mixcoatl',
    'mixcoatl.admin',
    'mixcoatl.analytics',
    'mixcoatl.automation',
    'mixcoatl.config',
    'mixcoatl.decorators',
    'mixcoatl.exceptions',
    'mixcoatl.geography',
    'mixcoatl.infrastructure',
    'mixcoatl.network',
    'mixcoatl.platform',
    'mixcoatl.settings'
]

requires = ['requests>=1.0.4', 'prettytable>=0.7.2', 'xmltodict>=0.4.2']

if sys.version_info < (2, 7):
    requires.append('argparse')

setup(
    name='mixcoatl',
    version=mixcoatl.__version__,
    description='Dell Cloud Manager API Python wrapper',
    long_description='Dell Cloud Manager API Python wrapper',
    author='Dell Cloud Management Team',
    author_email='mixcoatl@enstratius.com',
    url='https://github.com/enStratus/mixcoatl',
    packages=packages,
    package_data={'': ['LICENSE', 'README.rst', 'requirements.txt']},
    package_dir={'mixcoatl': 'mixcoatl'},
    include_package_data=True,
    install_requires=requires,
    scripts=glob.glob(os.path.join('bin','*')),
    license='Apache 2.0',
    zip_safe=False,
    classifiers=(
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7'
    ),
)



"""Implements the DCM Volume API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
from mixcoatl.infrastructure.snapshot import Snapshot
from mixcoatl.infrastructure.snapshot import SnapshotException
from mixcoatl.admin.job import Job
import json
import time


class Volume(Resource):

    """A volume is a block storage device that may be mounted by servers"""
    PATH = 'infrastructure/Volume'
    COLLECTION_NAME = 'volumes'
    PRIMARY_KEY = 'volume_id'

    def __init__(self, volume_id=None, endpoint=None, **kwargs):
        # pylint: disable-msg=W0613
        Resource.__init__(self, endpoint=endpoint)
        if 'detail' in kwargs:
            self.request_details = kwargs['detail']
        self.__volume_id = volume_id

    @property
    def volume_id(self):
        """`str` - The unique DCM id of the volume"""
        return self.__volume_id

    @lazy_property
    def array(self):
        """`dict` or `None` - The array, if any, that this volume belongs to"""
        return self.__array

    @lazy_property
    def available(self):
        """`bool` - Indicates if this volume is available for attachment"""
        return self.__available

    @lazy_property
    def customer(self):
        """`dict` - The customer to which this volume belongs"""
        return self.__customer

    @lazy_property
    def provider_id(self):
        """`str` - The unique cloud provider ID for this volume"""
        return self.__provider_id

    @lazy_property
    def size_string(self):
        """`str` - User-friendly string representing the size of the volume"""
        return self.__size_string

    @lazy_property
    def name(self):
        """`str` - A user-friendly name for the volume"""
        return self.__name

    @name.setter
    def name(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__name = b

    @lazy_property
    def label(self):
        """`str` - A user-friendly label for the volume"""
        return self.__label

    @label.setter
    def label(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__label = b

    @lazy_property
    def encrypted(self):
        """`bool` - Indicates if the volume is known by DCM as encrypted"""
        return self.__encrypted

    @lazy_property
    def region(self):
        """`dict` - The region in which this volume is stored"""
        return self.__region

    @lazy_property
    def budget(self):
        """`int` - The ID of the billing code against which costs are billed"""
        return self.__budget

    @budget.setter
    def budget(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__budget = b

    @lazy_property
    def server(self):
        """`dict` or `None` - The server to which this volume is attached if any"""
        return self.__server

    @lazy_property
    def owning_groups(self):
        """`list` - The groups who have ownership of this volume in DCM"""
        return self.__owning_groups

    @lazy_property
    def size_in_gb(self):
        """`int` - The size of the volume if reported by the cloud provider"""
        return self.__size_in_gb

    @size_in_gb.setter
    def size_in_gb(self, size):
        self.__size_in_gb = size

    @lazy_property
    def status(self):
        """`str` - The status of the volume in DCM"""
        return self.__status

    @lazy_property
    def removable(self):
        """`bool` - If this volume is removable or not"""
        return self.__removable

    @lazy_property
    def data_center(self):
        """`dict` - The data center where this volume is stored"""
        return self.__data_center

    @data_center.setter
    def data_center(self, dc):
        self.__data_center = {'data_center_id': dc}

    @lazy_property
    def owning_account(self):
        """`dict` - the DCM account under which this volume is registered"""
        return self.__owning_account

    @lazy_property
    def device_id(self):
        """`str` or `None` - The OS device id of this volume if attached"""
        return self.__device_id

    @lazy_property
    def creation_timestamp(self):
        """`str` - The timestamp when this volume was created"""
        return self.__creation_timestamp

    @lazy_property
    def owning_user(self):
        """`dict` - The DCM user who is the owner of this volume"""
        return self.__owning_user

    @lazy_property
    def cloud(self):
        """`dict` - The cloud where this volume is stored"""
        return self.__cloud

    @lazy_property
    def description(self):
        """`str` - The description of the volume in DCM"""
        return self.__description

    @description.setter
    def description(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__description = b

    @required_attrs(['volume_id'])
    def attach(self, server_id, device_id=None, callback=None):
        """Attach a volume to a server

        :param server_id: Server to attach the volume
        :type server_id: int.
        :param callback: Optional callback to send the results
        :returns: :class:`Volume`
        :raises: :class:`VolumeException`
        """
        payload = {'attach': [{'server': {'server_id': server_id}}]}

        if device_id is not None:
            payload['attach'][0]['device_id'] = device_id

        self.put(self.PATH + '/' + str(self.volume_id),
                 data=json.dumps(camel_keys(payload)))

        if self.last_error is None:
            if callback is None:
                return self
            else:
                callback(self)
        else:
            raise VolumeException(self.last_error)

    def detach(self, callback=None):
        """Detach a volume from a server

        :param callback: Optional callback to send the results
        :returns: :class:`Volume`
        :raises: :class:`VolumeException`
        """
        payload = '{"detach":[{}]}'
        path = self.PATH + '/' + str(self.volume_id)
        s = self.put(path, data=payload)
        if self.last_error is None:
            self.load()
            if callback is None:
                return s
            else:
                callback(s)
        else:
            raise VolumeException(self.last_error)

    @required_attrs(['budget', 'size_in_gb', 'description', 'data_center', 'name'])
    def create(self, callback=None):
        """Creates a new volume

        :param callback: Optional callback to send the resulting :class:`Job`
        :type callback: func.
        :returns: :class:`Job`
        :raises: :class:`VolumeCreationException`
        """
        optional_attrs = ['label']
        if self.volume_id is not None:
            raise VolumeCreationException(
                'Cannot create an already created volume: %s' % self.volume_id)

        parms = {'name': self.name, 'description': self.description, 'data_center': self.data_center,
                 'size_in_gb': self.size_in_gb, 'budget': self.budget}

        for oa in optional_attrs:
            try:
                if getattr(self, oa) is not None:
                    parms[0].update({oa: getattr(self, oa)})
            except AttributeError:
                pass

        payload = {'addVolume': [camel_keys(parms)]}
        self.post(data=json.dumps(payload))
        if self.last_error is None:
            j = Job(self.current_job, endpoint=self.endpoint)
            j.load()
            if callback is not None:
                callback(j)
            else:
                return j
        else:
            raise VolumeCreationException(self.last_error)

    def set_from_file(self, filename):
        """load server attributes from a json file
        :param filename: str with the full path to the json file containing the volume attributes
        """
        volume_file = open(filename).read()
        volume_dict = json.loads(volume_file)
        for key, value in volume_dict.items():
            setattr(self, key, value)

    def update(self):
        """Updates a volume with changed values

        :returns: `bool`
        :raises: :class:`VolumeException`
        """

        if self.pending_changes is None:
            pass
        elif 'budget' in self.pending_changes:
            self._change_budget()
        else:
            self._change_metadata()

    def destroy(self, reason=None):
        """Deletes the volume

        :param reason: Reason for removing the volume
        :type reason: str.
        :returns: `bool`
        :raises: :class:`VolumeException`
        """
        p = self.PATH + "/" + str(self.volume_id)
        qopts = {'reason': reason}
        return self.delete(p, params=qopts)

    def _change_metadata(self):
        """Changes metadata"""
        new_vals = {}
        for x in ['name', 'description', 'label']:
            if x in self.pending_changes:
                new_val = self.pending_changes[x]['new']
                new_vals[camel_keys(x)] = new_val
                self.pending_changes.pop(x, None)
        if len(new_vals) == 0:
            pass
        else:
            payload = {'describeVolume': [new_vals]}
            self.put(
                self.PATH + '/' + str(self.volume_id), data=json.dumps(payload))
            if self.last_error is None:
                self.load()
                return True
            else:
                raise VolumeException(self.last_error)

    @required_attrs(['budget'])
    def _change_budget(self):
        """Changes budget"""
        changes = self.pending_changes.pop('budget', None)
        if changes is None:
            pass
        else:
            payload = {'assignBudget': [{'budget': self.budget}]}
            self.put(
                self.PATH + '/' + str(self.volume_id), data=json.dumps(payload))
            if self.last_error is None:
                self.load()
                return True
            else:
                raise VolumeException(self.last_error)

    def snapshot(self, **kwargs):
        """Make a snapshot from a volume

        *This is not a part of the DCM API for Volume resources.
        It is a convenience wrapper around the Snapshot API from the
        perspective of an individual* :class:`Volume`

            .. warning::

                Snapshot creation is an asynchronous task.
                Specifying a callback will cause a blocking operation while the snapshot completes
                When using the callback, execution could block for a **VERY** long time.


        :param name: The name to assign the Snapshot. Default: `snap-<volume_id>-timestamp`
        :type name: str.
        :param description: Description of the snapshot. Default: `mixcoatl snapshot`
        :type description: str.
        :param budget: Budget to assign the snapshot. Default: :attr:`budget` of the current :class:`Volume`
        :type budget: int.
        :param callback: Optional callback to return the results.
        :type callback: func.
        :returns: :class:`Snapshot`
        :raises: :class:`VolumeSnapshotException`
        """
        tstamp = str(int(time.time()))
        if 'name' in kwargs:
            name = kwargs['name']
        else:
            name = 'snap-%s-%s' % (self.volume_id, tstamp)
        if 'description' in kwargs:
            description = kwargs['description']
        else:
            description = 'mixcoat snapshot'
        if 'budget' in kwargs:
            budget = kwargs['budget']
        else:
            budget = self.budget
        if 'callback' in kwargs:
            callback = kwargs['callback']
        else:
            callback = None
        try:
            return Snapshot.add_snapshot(self.volume_id,
                                         name,
                                         description,
                                         budget,
                                         callback=callback, endpoint=self.endpoint)
        except SnapshotException, e:
            raise VolumeSnapshotException(str(e))

    @classmethod
    def all(cls, endpoint=None, **kwargs):
        """List all volumes
        :param account_id: Restrict to volumes owned by `account_id`
        :type account_id: int.
        :param datacenter_id: Restrict to volumes based in `datacenter_id`
        :type datacenter_id: int.
        :param region_id: Restrict to volumes in `region_id`
        :type region_id: int.
        :param keys_only: Return :attr:`snapshot_id` or :class:`Snapshot`
        :type keys_only: bool.
        :param detail: Level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` of :attr:`volume_id` or :class:`Volume`
        :raises: :class:`VolumeException`
        """
        r = Resource(cls.PATH, endpoint=endpoint)
        params = {}

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        if 'datacenter_id' in kwargs:
            params['dataCenterId'] = kwargs['datacenter_id']
        if 'region_id' in kwargs:
            params['regionId'] = kwargs['region_id']
        if 'account_id' in kwargs:
            params['accountId'] = kwargs['account_id']

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise VolumeException(r.last_error)

    @classmethod
    def assign_budget(cls, volume_id, budget, endpoint=None):
        """Change the budget associated with a volume

        :param volume_id: The volume id to work with
        :type volume_id: int.
        :param budget: The budget code to assign
        :type budget: int.
        :param endpoint: The DCM endpoint
        :type endpoint: Endpoint
        :returns: `bool`
        :raises: :class:`VolumeException`
        """
        v = Volume(volume_id, endpoint=endpoint)
        v.budget = budget
        return v.update()

    @classmethod
    def assign_groups(cls, volume_id, group_id):
        """Change the group ownership of a volume

        :param volume_id: The volume id to work with
        :type volume_id: int.
        :param group_id: The group id to assign
        :type group_id: int.
        :returns: bool.
        :raises: :class:`VolumeException`
        """
        pass

    @classmethod
    def attach_volume(cls, volume_id, server_id, device_id=None, callback=None, endpoint=None):
        """Attach a volume to a server

            .. note::

                Attaching a volume is an asynchronous task.

        :param volume_id: The volume id to work with
        :type volume_id: int.
        :param server_id: The server to attach the volume
        :type server_id: int.
        :param device_id: Optional device id to assign the volume on the system
        :type device_id: str.
        :param callback: Optional callback to call with the results
        :type callback: func.
        :param endpoint: The DCM endpoint
        :type endpoint: Endpoint
        :returns: :class:`Job`
        :raises: :class:`VolumeException`
        """
        v = Volume(volume_id, endpoint=endpoint)
        v.request_details = 'basic'
        v.attach(server_id, device_id, callback)

    @classmethod
    def describe_volume(cls, volume_id, endpoint=None, **kwargs):
        """Change the DCM meta-data of a volume

        :param volume_id: The volume to modify
        :type volume_id: int.
        :param endpoint: The DCM endpoint
        :type endpoint: Endpoint
        :param description: Change the description
        :type description: str.
        :param name: Change the name
        :type name: str.
        :param label: Change the label. To remove the label, set to `None`
        :type label: str.
        :returns: :class:`Volume`
        :raises: :class:`VolumeException`
        """
        if kwargs is None:
            pass
        else:
            v = Volume(volume_id, endpoint=endpoint)
            v.request_details = 'basic'
            for attr in kwargs:
                setattr(v, attr, kwargs[attr])
            v.update()
            return v

    @classmethod
    def add_volume(cls, endpoint=None, **kwargs):
        """Create a new volume

        :param name: The name of the new volume
        :param size_in_gb: The size for the new volume
        :param data_center: The location of the new volume
        :param budget: The budget code for the new volume
        :param description: The description of the new volume
        :param callback: Optional callback to recieve the created Job
        :param endpoint: The DCM endpoint
        :type endpoint: Endpoint
        :returns: :class:`Job`
        :raises: :class:`VolumeCreationException`
        """
        v = Volume(endpoint=endpoint)
        cb = kwargs.pop('callback', None)

        for attr in kwargs:
            setattr(v, attr, kwargs[attr])

        v.create(callback=cb)

    @classmethod
    def detach_volume(cls, volume_id, callback=None, endpoint=None):
        """Detach a volume from a server

        :param volume_id: The volume to detach
        :type volume_id: int.
        :param callback: Optional callback to send the results
        :param endpoint: The DCM endpoint
        :type endpoint: Endpoint
        :returns: :class:`Volume`
        :raises: :class:`VolumeException`
        """
        v = Volume(volume_id)
        v.detach(callback=callback)


class VolumeException(BaseException):

    """Generic Volume Exception"""
    pass


class VolumeSnapshotException(VolumeException):

    """Volume Snapshot Exception"""
    pass


class VolumeCreationException(VolumeException):

    """Volume Creation Exception"""
    pass



"""Implements access to the DCM Snapshot API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
from mixcoatl.admin.job import Job
import json


class Snapshot(Resource):

    """A snapshot is a point-in-time snapshot of a volume"""
    PATH = 'infrastructure/Snapshot'
    COLLECTION_NAME = 'snapshots'
    PRIMARY_KEY = 'snapshot_id'

    def __init__(self, snapshot_id=None, endpoint=None, *args, **kwargs):
        # pylint: disable-msg=W0613
        Resource.__init__(self, endpoint=endpoint)
        self.__snapshot_id = snapshot_id

    @property
    def snapshot_id(self):
        """`int` - The unique DCM id for this snapshot"""
        return self.__snapshot_id

    @lazy_property
    def available(self):
        """`bool` - Identifies if the snapshot is available for creating volumes"""
        return self.__available

    @lazy_property
    def budget(self):
        """`int` - The id of the billing code against which costs will be associated"""
        # pylint: disable-msg=E0202
        return self.__budget

    @budget.setter
    def budget(self, budget):
        # pylint: disable-msg=C0111,W0201
        self.__budget = budget

    @lazy_property
    def cloud(self):
        """`dict` - The cloud in which this snapshot is stored"""
        return self.__cloud

    @lazy_property
    def created_timestamp(self):
        """`str` - The timestamp when the snapshot was created"""
        return self.__created_timestamp

    @lazy_property
    def customer(self):
        """`dict` - The customer to which the snapshot belongs"""
        return self.__customer

    @lazy_property
    def description(self):
        """`str` - The description of the snapshot established in DCM"""
        # pylint: disable-msg=E0202
        return self.__description

    @description.setter
    def description(self, desc):
        # pylint: disable-msg=C0111,W0201
        self.__description = desc

    @lazy_property
    def encrypted(self):
        """`bool` - Is the snapshot known to be encrypted by DCM"""
        return self.__encrypted

    @lazy_property
    def name(self):
        """`str` - The user-friendly name of the snapshot"""
        # pylint: disable-msg=E0202
        return self.__name

    @name.setter
    def name(self, name):
        # pylint: disable-msg=C0111,W0201
        self.__name = name

    @lazy_property
    def label(self):
        """`str` - The label assigned to the snapshot"""
        return self.__label

    @label.setter
    def label(self, label):
        # pylint: disable-msg=C0111,W0201
        self.__label = label

    @lazy_property
    def owning_account(self):
        """`dict` or `None` - The DCM account where the snapshot is registered"""
        return self.__owning_account

    @lazy_property
    def owning_user(self):
        """`dict` or `None` - The DCM user who has ownership of this snapshot"""
        return self.__owning_user

    @lazy_property
    def owning_groups(self):
        """`dict` - The DCM groups who have ownership of this snapshot"""
        return self.__owning_groups

    @lazy_property
    def provider_id(self):
        """`str` - The cloud provider's unique id for the snapshot"""
        return self.__provider_id

    @lazy_property
    def region(self):
        """`dict` - The cloud region where the snapshot is stored"""
        return self.__region

    @lazy_property
    def removable(self):
        """`bool` - Indicates if this snapshot can be removed"""
        return self.__removable

    @lazy_property
    def sharable(self):
        """`bool` - Indicates if this snapshot can be shared"""
        return self.__sharable

    @lazy_property
    def size_in_gb(self):
        """`int` - The size of the snapshot in GB"""
        return self.__size_in_gb

    @lazy_property
    def status(self):
        """`str` - The DCM status of the snapshot *(`ACTIVE`|`INACTIVE`)*"""
        return self.__status

    @lazy_property
    def volume(self):
        """`dict` or `None` - The volume, if known, from which the snapshot was created"""
        # pylint: disable-msg=E0202
        return self.__volume

    @volume.setter
    def volume(self, volume_id):
        # pylint: disable-msg=C0111,W0201
        self.__volume = {'volume_id': volume_id}

    @required_attrs(['snapshot_id'])
    def destroy(self, reason='No reason given'):
        """delete a snapshot

        :param reason: A reason for deleting the snapshot
        :type reason: str.
        :returns: `bool`
        :raises: :class:`SnapshotException`
        """
        params = {'reason': reason}

        try:
            return self.delete(self.PATH + '/' + str(self.snapshot_id), params=params)
        except:
            raise SnapshotException(self.last_error)

    def update(self):
        """Updates a snapshot with changed values

        :returns: :class:`Snapshot`
        :raises: :class:`SnapshotException`
        """

        if self.pending_changes is None:
            pass
        else:
            payload = {'describeSnapshot': [{}]}
            for x in ['name', 'description', 'label']:
                if x in self.pending_changes:
                    new_val = self.pending_changes[x]['new']
                    payload['describeSnapshot'][0][camel_keys(x)] = new_val
                    self.pending_changes.pop(x, None)
            if len(payload['describeSnapshot'][0]) == 0:
                pass
            else:
                self.put(
                    self.PATH + '/' + str(self.snapshot_id), data=json.dumps(payload))

            if self.last_error is None:
                self.load()
                return self
            else:
                raise SnapshotException(self.last_error)

    @required_attrs(['volume', 'name', 'description', 'budget'])
    def create(self, callback=None):
        """Creates a snapshot

        :returns: :class:`Snapshot`
        :raises: :class:`SnapshotException`
        """

        if self.snapshot_id is not None:
            raise SnapshotException(
                'Cannot snapshot a snapshot: %s' % self.snapshot_id)

        payload = {'addSnapshot': [{}]}
        payload['addSnapshot'][0]['volume'] = camel_keys(self.volume)
        payload['addSnapshot'][0]['name'] = self.name
        payload['addSnapshot'][0]['description'] = self.description
        payload['addSnapshot'][0]['budget'] = self.budget
        optional_attrs = ['label']

        for oa in optional_attrs:
            try:
                if getattr(self, oa) is not None:
                    payload['addSnapshot'][0].update(
                        camel_keys({oa: getattr(self, oa)}))
            except AttributeError:
                # We did say optional....
                pass
        self.post(self.PATH, data=json.dumps(payload))
        if self.last_error is None:
            if callback is not None:
                callback(self)
            else:
                return self
        else:
            raise SnapshotException(self.last_error)

    @classmethod
    def all(cls, endpoint=None, **kwargs):
        """Return a list of snapshots

        :param account_id: Restrict to snapshots owned by `account_id`
        :type account_id: int.
        :param volume_id: Restrict to snapshots based on `volume_id`
        :type volume_id: int.
        :param region_id: Restrict to snapshots in `region_id`
        :type region_id: int.
        :param keys_only: Return :attr:`snapshot_id` or :class:`Snapshot`
        :type keys_only: bool.
        :param detail: Level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` of :attr:`snapshot_id` or :class:`Snapshot`
        :raises: :class:`SnapshotException`
        """
        r = Resource(cls.PATH, endpoint=endpoint)
        params = {}

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        if 'region_id' in kwargs:
            params['regionId'] = kwargs['region_id']
        if 'account_id' in kwargs:
            params['accountId'] = kwargs['account_id']
        if 'volume_id' in kwargs:
            params['volumeId'] = kwargs['volume_id']

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise SnapshotException(r.last_error)

    @classmethod
    def describe_snapshot(cls, snapshot_id, endpoint=None, **kwargs):
        """Changes the basic metadata for a snapshot

        :param id: The snapshot to modify
        :type id: int.
        :param description: Change the description.
        :type description: str.
        :param name: Change the name.
        :type name: str.
        :param label: Change the label. To remove the label, set to `None`
        :type label: str.
        :returns: :class:`Snapshot`
        :raises: :class:`SnapshotException`
        """
        s = cls(snapshot_id, endpoint=endpoint)
        for x in ['name', 'description', 'label']:
            if x in kwargs:
                setattr(s, x, kwargs[x])
        s.update()
        return s

    @classmethod
    def delete_snapshot(cls, snapshot_id, reason, endpoint=None):
        """delete a snapshot

        :param snapshot_id: The DCM snapshot id
        :type snapshot_id: int
        :param reason: A reason for deleting the snapshot
        :type reason: str.
        :returns: `bool`
        :raises: :class:`SnapshotException`
        """
        s = cls(snapshot_id, endpoint=endpoint)
        return s.destroy(reason=reason)

    @classmethod
    def add_snapshot(cls, volume_id, name, description, budget, callback=None, endpoint=None):
        """Creates a snapshot from `volume_id`

            .. warning::

                Snapshot creation is an asynchronous task.
                Specifying a callback will cause a blocking operation while the snapshot completes.
                When using the callback, execution could block for a **VERY** long time depending on the time it takes to make the snapshot.

        :param volume_id: The volume to snapshot
        :type volume_id: int.
        :param name: The name for the snapshot
        :type name: str.
        :param description: Description of the snapshot
        :type description: str.
        :param budget: The billing code for the snapshot
        :type budget: int.
        :param callback: An optional callback to send the final :class:`Snapshot`.
        :type callback: func.
        :returns: :class:`Snapshot`
        :raises: :class:`SnapshotException`
        """

        s = cls()
        s.volume = volume_id
        s.name = name
        s.budget = budget
        s.description = description
        s.create()
        if s.current_job is None:
            raise SnapshotException('No job found. This is...odd')
        else:
            if callback is not None:
                job = Job.wait_for(s.current_job)
                if job is True:
                    try:
                        j = Job(s.current_job, endpoint=endpoint)
                        snapshot = cls(j.message)
                        snapshot.load()
                        callback(snapshot)
                    except:
                        raise SnapshotException("Unhandled error in callback")
            else:
                return s


class SnapshotException(BaseException):

    """Generic exception for Snapshots"""
    pass



"""
mixcoatl.admin.customer
--------------------
"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys, uncamel_keys
import json


class Customer(Resource):

    """The customer resource represents the overarching Dell Cloud Manager customer account. A
    customer may have any number of actual accounts associated with them. Each account is
    billed separately, but share customer-wide resources like users, billing codes, standard networks,
    standard ports, and more."""
    PATH = 'admin/Customer'
    COLLECTION_NAME = 'customers'
    PRIMARY_KEY = 'customer_id'

    def __init__(self, role_id=None, endpoint=None, *args, **kwargs):
        Resource.__init__(self, endpoint=endpoint)
        self.__customer_id = customer_id

    @property
    def customer_id(self):
        """`int` - The unique id of the customer"""
        return self.__customer_id

    @classmethod
    def all(cls, keys_only=False, endpoint=None, **kwargs):
        """Get all customers"""
        r = Resource(cls.PATH, endpoint=endpoint)
        params = {}

        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        if 'customer_id' in kwargs:
            params['customer_id'] = kwargs['customer_id']

        x = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [type(cls.__name__, (object,), i) for i in uncamel_keys(x)[cls.COLLECTION_NAME]]
        else:
            raise CustomerException(r.last_error)


class CustomerException(BaseException):
    pass
